from __future__ import annotations

from optuna_dashboard.preferential._study import create_study
from optuna_dashboard.preferential._study import load_study
from optuna_dashboard.preferential._study import PreferentialStudy


__all__ = [
    "PreferentialStudy",
    "create_study",
    "load_study",
]



from __future__ import annotations

import threading

from bottle import Bottle
from bottle import redirect
from bottle import request
from bottle import SimpleTemplate
from optuna.storages import BaseStorage
from optuna.storages import RDBStorage
from optuna.version import __version__ as optuna_ver

from ._bottle_util import BottleViewReturn


rdb_schema_migrate_lock = threading.Lock()
rdb_schema_needs_migrate = False
rdb_schema_unsupported = False
rdb_schema_template = SimpleTemplate(
    """<!DOCTYPE html>
<html lang="en">
<head>
<title>Incompatible RDB Schema Error - Optuna Dashboard</title>
<meta charset="UTF-8" />
<meta name="viewport" content="width=device-width, initial-scale=1" />
<style>
body {
    padding: 0;
    height: 100vh;
    display: flex;
    flex-direction: column;
    justify-content: center;
    align-items: center;
}
.wrapper {
    padding: 64px;
    width: 600px;
    background-color: rgb(255, 255, 255);
    box-shadow: rgba(0, 0, 0, 0.08) 0 8px 24px;
    margin: 0px auto;
    border-radius: 8px;
}
</style>
</head>
<body>
    <div class="wrapper">
    <h1>Error: Incompatible RDB Schema</h1>
% if rdb_schema_unsupported:
    <p>Your Optuna version {{ optuna_ver }} seems outdated against the storage version. Please try updating optuna to the latest version by `$ pip install -U optuna`.</p>
% elif rdb_schema_needs_migrate:
    <p>The runtime optuna version {{ optuna_ver }} is no longer compatible with the table schema. Please execute `$ optuna storage upgrade --storage $STORAGE_URL` or press the following button for upgrading the storage.</p>
    <form action="/incompatible-rdb-schema" method="post">
    <button>Migrate</button>
    </form>
% end
    </div>
</body>
</html>"""  # noqa: E501
)


def update_schema_compatibility_flags(storage: RDBStorage) -> None:
    global rdb_schema_needs_migrate, rdb_schema_unsupported

    with rdb_schema_migrate_lock:
        current_version = storage.get_current_version()
        head_version = storage.get_head_version()
        rdb_schema_needs_migrate = current_version != head_version
        rdb_schema_unsupported = current_version not in storage.get_all_versions()


def is_incompatible() -> bool:
    return rdb_schema_needs_migrate or rdb_schema_unsupported


def register_rdb_migration_route(app: Bottle, storage: BaseStorage) -> None:
    if isinstance(storage, RDBStorage):
        update_schema_compatibility_flags(storage)

    @app.get("/incompatible-rdb-schema")
    def get_incompatible_rdb_schema() -> BottleViewReturn:
        if not is_incompatible() or not isinstance(storage, RDBStorage):
            return redirect("/dashboard", 302)
        return rdb_schema_template.render(
            rdb_schema_needs_migrate=rdb_schema_needs_migrate,
            rdb_schema_unsupported=rdb_schema_unsupported,
            optuna_ver=optuna_ver,
        )

    @app.post("/incompatible-rdb-schema")
    def post_incompatible_rdb_schema() -> BottleViewReturn:
        if not isinstance(storage, RDBStorage):
            return redirect("/dashboard", 302)

        global rdb_schema_needs_migrate
        assert not rdb_schema_unsupported
        with rdb_schema_migrate_lock:
            storage.upgrade()
            rdb_schema_needs_migrate = False
        return redirect("/dashboard", 302)

    @app.hook("before_request")
    def check_schema_compatibility() -> None:
        if not isinstance(storage, RDBStorage):
            return

        if request.path != "/" and not request.path.startswith("/dashboard"):
            return

        update_schema_compatibility_flags(storage)
        if is_incompatible():
            return redirect("/incompatible-rdb-schema", 302)



# This code is taken from the below to keep support for older versions.
# https://github.com/optuna/optuna/blob/v3.0.5/optuna/study/_multi_objective.py

from __future__ import annotations

from typing import Optional
from typing import Sequence

from optuna.study._study_direction import StudyDirection
from optuna.trial import FrozenTrial
from optuna.trial import TrialState


def get_pareto_front_trials(
    trials: Sequence[FrozenTrial], directions: Sequence[StudyDirection]
) -> list[FrozenTrial]:
    if len(directions) == 2:
        return _get_pareto_front_trials_2d(trials, directions)  # Log-linear in number of trials.
    return _get_pareto_front_trials_nd(trials, directions)  # Quadratic in number of trials.


def _get_pareto_front_trials_2d(
    trials: Sequence[FrozenTrial], directions: Sequence[StudyDirection]
) -> list[FrozenTrial]:
    trials = [trial for trial in trials if trial.state == TrialState.COMPLETE]

    n_trials = len(trials)
    if n_trials == 0:
        return []

    trials.sort(
        key=lambda trial: (
            _normalize_value(trial.values[0], directions[0]),
            _normalize_value(trial.values[1], directions[1]),
        ),
    )

    last_nondominated_trial = trials[0]
    pareto_front = [last_nondominated_trial]
    for i in range(1, n_trials):
        trial = trials[i]
        if _dominates(last_nondominated_trial, trial, directions):
            continue
        pareto_front.append(trial)
        last_nondominated_trial = trial

    pareto_front.sort(key=lambda trial: trial.number)
    return pareto_front


def _get_pareto_front_trials_nd(
    trials: Sequence[FrozenTrial], directions: Sequence[StudyDirection]
) -> list[FrozenTrial]:
    pareto_front = []
    trials = [t for t in trials if t.state == TrialState.COMPLETE]

    # TODO(vincent): Optimize (use the fast non dominated sort defined in the NSGA-II paper).
    for trial in trials:
        dominated = False
        for other in trials:
            if _dominates(other, trial, directions):
                dominated = True
                break

        if not dominated:
            pareto_front.append(trial)

    return pareto_front


def _dominates(
    trial0: FrozenTrial, trial1: FrozenTrial, directions: Sequence[StudyDirection]
) -> bool:
    values0 = trial0.values
    values1 = trial1.values

    assert values0 is not None
    assert values1 is not None

    if len(values0) != len(values1):
        raise ValueError("Trials with different numbers of objectives cannot be compared.")

    if len(values0) != len(directions):
        raise ValueError(
            "The number of the values and the number of the objectives are mismatched."
        )

    if trial0.state != TrialState.COMPLETE:
        return False

    if trial1.state != TrialState.COMPLETE:
        return True

    normalized_values0 = [_normalize_value(v, d) for v, d in zip(values0, directions)]
    normalized_values1 = [_normalize_value(v, d) for v, d in zip(values1, directions)]

    if normalized_values0 == normalized_values1:
        return False

    return all(v0 <= v1 for v0, v1 in zip(normalized_values0, normalized_values1))


def _normalize_value(value: Optional[float], direction: StudyDirection) -> float:
    if value is None:
        value = float("inf")

    if direction is StudyDirection.MAXIMIZE:
        value = -value

    return value



from __future__ import annotations

from typing import TYPE_CHECKING

from optuna_dashboard.artifact.exceptions import ArtifactNotFound as DashboardArtifactNotFound


if TYPE_CHECKING:
    from typing import BinaryIO
    from typing import TypeGuard

    from optuna.artifacts._protocol import ArtifactStore

    from .protocol import ArtifactBackend


def is_artifact_backend(store: ArtifactBackend | ArtifactStore) -> TypeGuard[ArtifactBackend]:
    return getattr(store, "open_reader", None) is None


def to_artifact_store(store: ArtifactBackend | ArtifactStore) -> ArtifactStore:
    if is_artifact_backend(store):
        return ArtifactBackendToStore(store)
    # mypy cannot infer the type of `store` here.
    return store  # type: ignore


class ArtifactBackendToStore:
    """Converts a Dashboard's ArtifactBackend to Optuna's ArtifactStore."""

    def __init__(self, artifact_backend: ArtifactBackend) -> None:
        self._backend = artifact_backend

    def open_reader(self, artifact_id: str) -> BinaryIO:
        from optuna.artifacts.exceptions import ArtifactNotFound

        try:
            return self._backend.open(artifact_id)
        except DashboardArtifactNotFound as e:
            raise ArtifactNotFound from e

    def write(self, artifact_id: str, content_body: BinaryIO) -> None:
        self._backend.write(artifact_id, content_body)

    def remove(self, artifact_id: str) -> None:
        from optuna.artifacts.exceptions import ArtifactNotFound

        try:
            self._backend.remove(artifact_id)
        except DashboardArtifactNotFound as e:
            raise ArtifactNotFound from e



from __future__ import annotations

import copy
from typing import Any
from typing import Container
from typing import Iterable

import optuna
from optuna import logging
from optuna._imports import try_import
from optuna.distributions import BaseDistribution
from optuna.samplers import BaseSampler
from optuna.trial import FrozenTrial
from optuna.trial import TrialState
from optuna_dashboard.preferential._system_attrs import get_n_generate
from optuna_dashboard.preferential._system_attrs import get_preferences
from optuna_dashboard.preferential._system_attrs import get_skipped_trial_ids
from optuna_dashboard.preferential._system_attrs import is_skipped_trial
from optuna_dashboard.preferential._system_attrs import report_preferences
from optuna_dashboard.preferential._system_attrs import set_n_generate


with try_import() as _imports:
    from optuna_dashboard.preferential.samplers.gp import PreferentialGPSampler


_logger = logging.get_logger(__name__)
_SYSTEM_ATTR_PREFERENTIAL_STUDY = "preference:is_preferential"


class PreferentialStudy:
    """A Study-like class for preferential optimization.

    This object provides interfaces to create a new `Trial`_, set/get results
    of pairwise comparison called preferences.

    .. _Trial: https://optuna.readthedocs.io/en/stable/reference/generated/\
    optuna.trial.Trial.html#optuna.trial.Trial

    Note that the direct use of this constructor is not recommended.
    To create and load a study, please refer to the documentation of
    :func:`~optuna_dashboard.preferential.create_study` and
    :func:`~optuna_dashboard.preferential.load_study` respectively.

    .. note::
        Preferential optimization is an experimental feature (introduced in v0.13.0).
        The interface may change in newer versions without prior notice.
    """

    def __init__(self, study: optuna.Study) -> None:
        self._study = study

    @property
    def trials(self) -> list[FrozenTrial]:
        """Return the all trials.

        .. seealso::

            See `Study.trials`_ for details.

            .. _Study.trials: https://optuna.readthedocs.io/en/stable/reference/generated/\
            optuna.study.Study.html#optuna.study.Study.trials

        Returns:
            A list of FrozenTrial object
        """
        return self._study.trials

    @property
    def best_trials(self) -> list[FrozenTrial]:
        """Return the trials that is not dominated by other trials.

        Returns:
            A list of FrozenTrial object
        """
        return get_best_trials(self._study._study_id, self._study._storage)

    @property
    def study_name(self) -> str:
        """Return the name of the study.

        Returns:
            A string object
        """
        return self._study.study_name

    @property
    def user_attrs(self) -> dict[str, Any]:
        """Return user attributes of the study.

        .. seealso::

            See `Study.user_attrs`_ for details.

            .. _Study.user_attrs: https://optuna.readthedocs.io/en/stable/reference/\
            generated/optuna.study.Study.html#optuna.study.Study.user_attrs

        Returns:
            A dictionary containing all user attributes
        """
        return self._study.user_attrs

    @property
    def preferences(self) -> list[tuple[FrozenTrial, FrozenTrial]]:
        """Return results of pairwise comparison.

        Returns:
            A list of the pair of FrozenTrial objects. The left trial is better than the right one.
        """
        return self.get_preferences(deepcopy=True)

    def get_trials(
        self,
        deepcopy: bool = True,
        states: Container[optuna.trial.TrialState] | None = None,
    ) -> list[FrozenTrial]:
        """Return the trials that is not dominated by other trials.

        .. seealso::

            See `Study.get_trials`_ for details.

            .. _Study.get_trials: https://optuna.readthedocs.io/en/stable/reference/\
            generated/optuna.study.Study.html#optuna.study.Study.get_trials

        Args:
            deepcopy:
                Flag to control whether to apply ``copy.deepcopy()`` to the trials.
                Note that if you set the flag to :obj:`False`, you shouldn't mutate
                any fields of the returned trial. Otherwise the internal state of
                the study may corrupt and unexpected behavior may happen.
            states:
                Trial states to filter on. If :obj:`None`, include all states.

        Returns:
            A list of FrozenTrial object
        """
        return self._study.get_trials(deepcopy, states)

    def ask(self, fixed_distributions: dict[str, BaseDistribution] | None = None) -> optuna.Trial:
        """Create a new trial from which hyperparameters can be suggested.

        .. seealso::

            See `Study.ask`_ for details.

            .. _Study.ask: https://optuna.readthedocs.io/en/stable/reference/\
            generated/optuna.study.Study.html#optuna.study.Study.ask

        Args:
            fixed_distributions:
                A dictionary containing the parameter names and parameter's distributions. Each
                parameter in this dictionary is automatically suggested for the returned trial,
                even when the suggest method is not explicitly invoked by the user. If this
                argument is set to :obj:`None`, no parameter is automatically suggested.

        Returns:
            A Trial object.
        """
        return self._study.ask(fixed_distributions)

    def add_trial(self, trial: FrozenTrial) -> None:
        """Add a trial to the study.

        .. seealso::

            See `Study.add_trials()`_ for details.

            .. _Study.add_trials(): https://optuna.readthedocs.io/en/stable/reference/\
            generated/optuna.study.Study.html#optuna.study.Study.add_trials
        """
        self._study.add_trial(trial)

    def add_trials(self, trials: Iterable[FrozenTrial]) -> None:
        """Add trials to the study.

        .. seealso::

            See `Study.add_trials()`_ for details.

            .. _Study.add_trials(): https://optuna.readthedocs.io/en/stable/reference/\
            generated/optuna.study.Study.html#optuna.study.Study.add_trials
        """
        self._study.add_trials(trials)

    def enqueue_trial(
        self,
        params: dict[str, Any],
        user_attrs: dict[str, Any] | None = None,
        skip_if_exists: bool = False,
    ) -> None:
        """Enqueue a trial with given parameter values.

        You can fix the next sampling parameters which will be evaluated in your
        objective function.

        .. seealso::

            See `Study.enqueue_trials`_ for details.

            .. _Study.get_trials: https://optuna.readthedocs.io/en/stable/reference/\
            generated/optuna.study.Study.html#optuna.study.Study.enqueue_trials

        Args:
            params:
                Parameter values to pass your objective function.
            user_attrs:
                A dictionary of user-specific attributes other than ``params``.
            skip_if_exists:
                When :obj:`True`, prevents duplicate trials from being enqueued again.

                .. note::
                    This method might produce duplicated trials if called simultaneously
                    by multiple processes at the same time with same ``params`` dict.
        """
        self._study.enqueue_trial(params, user_attrs, skip_if_exists)

    def report_preference(
        self,
        better_trials: FrozenTrial | list[FrozenTrial],
        worse_trials: FrozenTrial | list[FrozenTrial],
    ) -> None:
        """Report results of pairwise comparison.

        Args:
            better_trials:
                Trials that are better than worse_trials.
            worse_trials:
                Trials that are worse than better_trials.
        """
        if not isinstance(better_trials, list):
            better_trials = [better_trials]
        if not isinstance(worse_trials, list):
            worse_trials = [worse_trials]

        report_preferences(
            self._study._study_id,
            self._study._storage,
            [(b.number, w.number) for b in better_trials for w in worse_trials],
        )

    def get_preferences(self, *, deepcopy: bool = True) -> list[tuple[FrozenTrial, FrozenTrial]]:
        """Return results of pairwise comparison.

        Args:
            deepcopy:
                Flag to control whether to apply ``copy.deepcopy()`` to the trials.
                Note that if you set the flag to :obj:`False`, you shouldn't mutate
                any fields of the returned trial. Otherwise the internal state of
                the study may corrupt and unexpected behavior may happen.

        Returns:
            A list of the pair of FrozenTrial objects. The left trial is better than the right one.
        """

        preferences = get_preferences(
            self._study._storage.get_study_system_attrs(self._study._study_id)
        )  # Must come before study.get_trials()
        trials = self._study.get_trials(deepcopy=deepcopy)
        return [(trials[better], trials[worse]) for (better, worse) in preferences]

    def set_user_attr(self, key: str, value: Any) -> None:
        """Set a user attribute to the study.

        Args:
            key: A key string of the attribute.
            value: A value of the attribute. The value should be JSON serializable.

        .. seealso::

            See the `tutorial for user attributes <https://optuna.readthedocs.io/en/stable/\
            tutorial/20_recipes/003_attributes.html>`_ on Optuna's documentation.
        """
        self._study.set_user_attr(key, value)

    def should_generate(self) -> bool:
        """Return whether the generator should generate a new trial now.

        Returns :obj:`True` if the number of trials not reported bad and not skipped are less than
        :attr:`~optuna_dashboard.preferential.PreferentialStudy.n_generate`. Users are recommended
        to generate a new trial if this method returns :obj:`True`, and to wait for human
        evaluation if this method returns :obj:`False`.
        """
        study_system_attrs = self._study._storage.get_study_system_attrs(
            self._study._study_id
        )  # Must come before _study.get_trials()
        trials = self._study.get_trials(
            deepcopy=False, states=(TrialState.COMPLETE, TrialState.RUNNING)
        )
        worse_trial_numbers = {worse for _, worse in get_preferences(study_system_attrs)}
        skipped_trial_ids = set(get_skipped_trial_ids(study_system_attrs))
        active_trials = [
            t
            for t in trials
            if t.number not in worse_trial_numbers and t._trial_id not in skipped_trial_ids
        ]
        return len(active_trials) < get_n_generate(
            self._study._storage.get_study_system_attrs(self._study._study_id)
        )


def get_best_trials(study_id: int, storage: optuna.storages.BaseStorage) -> list[FrozenTrial]:
    preferences = get_preferences(storage.get_study_system_attrs(study_id))
    worse_numbers = {worse for _, worse in preferences}
    nondominated_numbers = {better for better, _ in preferences if better not in worse_numbers}
    trials = storage.get_all_trials(study_id, deepcopy=False)

    study_system_attrs = storage.get_study_system_attrs(study_id)

    best_trials = []
    for n in nondominated_numbers:
        t = trials[n]
        if is_skipped_trial(t._trial_id, study_system_attrs):
            continue
        best_trials.append(copy.deepcopy(t))
    return best_trials


def create_study(
    *,
    n_generate: int,
    storage: str | optuna.storages.BaseStorage | None = None,
    sampler: BaseSampler | None = None,
    study_name: str | None = None,
    load_if_exists: bool = False,
) -> PreferentialStudy:
    """Like ``optuna.create_study()``, but for preferential optimization.

    Example:

        .. testcode::

            import optuna
            from optuna_dashboard.preferential import create_study


            study = create_study()
            trial = study.ask()

    Args:
        n_generate:
            The number of active trials to keep.
            :func:`~optuna_dashboard.preferential.PreferentialStudy.should_generate` returns
            :obj:`True` if the number of trials not reported bad and not skipped are less than
            ``n_generate``.

        storage:
            Database URL. If this argument is set to None, in-memory storage is used, and the
            :class:`~optuna_dashboard.preferential.PreferentialStudy` will not be persistent.

        sampler:
            A sampler object that implements background algorithm for value suggestion.
            If :obj:`None` is specified,
            :class:`~optuna_dashboard.preferential.samplers.gp.PreferentialGPSampler` is used.
            Please note that most Optuna samplers does not work efficiently for preferential
            optimization.

        study_name:
            Study's name. If this argument is set to None, a unique name is generated
            automatically.

        load_if_exists:
            Flag to control the behavior to handle a conflict of study names.
            In the case where a study named ``study_name`` already exists in the ``storage``,
            a :class:`~optuna.exceptions.DuplicatedStudyError` is raised if ``load_if_exists`` is
            set to :obj:`False`.
            Otherwise, the creation of the study is skipped, and the existing one is returned.

    Returns:
        A :class:`~optuna_dashboard.preferential.PreferentialStudy` object.

    .. note::
        Preferential optimization is an experimental feature (introduced in v0.13.0).
        The interface may change in newer versions without prior notice.
    """
    try:
        if sampler is None:
            _imports.check()  # If BoTorch is not installed, raise ImportError.
            sampler = PreferentialGPSampler()

        study = optuna.create_study(
            storage=storage,
            sampler=sampler,
            study_name=study_name,
        )
        study._storage.set_study_system_attr(
            study._study_id, _SYSTEM_ATTR_PREFERENTIAL_STUDY, True
        )
        set_n_generate(study._study_id, study._storage, n_generate)
        return PreferentialStudy(study)

    except optuna.exceptions.DuplicatedStudyError:
        if load_if_exists:
            assert study_name is not None
            assert storage is not None

            _logger.info(
                "Using an existing study with name '{}' instead of creating a new one.".format(
                    study_name
                )
            )
            return load_study(
                study_name=study_name,
                storage=storage,
                sampler=sampler,
            )
        else:
            raise


def load_study(
    *,
    study_name: str | None,
    storage: str | optuna.storages.BaseStorage,
    sampler: BaseSampler | None = None,
) -> PreferentialStudy:
    """Like ``optuna.load_study()``, but for preferential optimization.

    Example:

        .. testsetup::

            import os

            if os.path.exists("example.db"):
                raise RuntimeError("'example.db' already exists. Please remove it.")

        .. testcode::

            import optuna
            from optuna_dashboard.preferential import create_study
            from optuna_dashboard.preferential import load_study

            study = create_study(storage="sqlite:///example.db", study_name="my_study")
            study.ask()

            loaded_study = load_study(study_name="my_study", storage="sqlite:///example.db")
            assert len(loaded_study.trials) == len(study.trials)

        .. testcleanup::

            os.remove("example.db")

    Args:
        study_name:
            Study's name. Each study has a unique name as an identifier. If :obj:`None`, checks
            whether the storage contains a single study, and if so loads that study.
            ``study_name`` is required if there are multiple studies in the storage.
        storage:
            Database URL such as ``sqlite:///example.db``. Please see also the documentation of
            :func:`~optuna.study.create_study` for further details.
        sampler:
            A sampler object that implements background algorithm for value suggestion.
            If :obj:`None` is specified,
            :class:`~optuna_dashboard.preferential.samplers.gp.PreferentialGPSampler` is used.
            Please note that most Optuna samplers does not work efficiently for preferential
            optimization.

    Returns:
        A :class:`~optuna_dashboard.preferential.PreferentialStudy` object.

    .. note::
        Preferential optimization is an experimental feature (introduced in v0.13.0).
        The interface may change in newer versions without prior notice.
    """
    if sampler is None:
        _imports.check()  # If BoTorch is not installed, raise ImportError.
        sampler = PreferentialGPSampler()

    study = optuna.load_study(study_name=study_name, storage=storage, sampler=sampler)
    system_attrs = study._storage.get_study_system_attrs(study._study_id)
    if not system_attrs.get(_SYSTEM_ATTR_PREFERENTIAL_STUDY):
        raise ValueError("The study is not a PreferentialStudy.")
    return PreferentialStudy(study)



from __future__ import annotations

import itertools
import math
from typing import Any
from typing import Callable
import warnings

import botorch.acquisition.analytic
import botorch.models.model
import botorch.optim
import botorch.posteriors.gpytorch
import gpytorch.constraints
import gpytorch.kernels
from gpytorch.likelihoods.gaussian_likelihood import Prior
import numpy as np
import optuna
import optuna._transform
from optuna.distributions import CategoricalDistribution
from optuna.distributions import FloatDistribution
from optuna.distributions import IntDistribution
import torch
from torch import Tensor

from .._system_attrs import get_preferences


def _orthants_MVN_Gibbs_sampling(cov_inv: Tensor, cycles: int, initial_sample: Tensor) -> Tensor:
    dim = cov_inv.shape[0]
    assert cov_inv.shape == (dim, dim)

    sample_chain = initial_sample
    conditional_std = torch.rsqrt(torch.diag(cov_inv))
    scaled_cov_inv = cov_inv / torch.diag(cov_inv)[:, None]

    out = torch.empty((cycles + 1, dim), dtype=torch.float64)
    out[0, :] = sample_chain

    for i in range(cycles):
        for j in range(dim):
            conditional_mean = sample_chain[j] - scaled_cov_inv[j] @ sample_chain
            sample_chain[j] = (
                _one_side_trunc_norm_sampling(lower=-conditional_mean / conditional_std[j])
                * conditional_std[j]
                + conditional_mean
            )
        out[i + 1, :] = sample_chain

    return out


def _one_side_trunc_norm_sampling(lower: Tensor) -> Tensor:
    r = torch.rand(torch.Size(()), dtype=torch.float64)
    ret = -torch.special.ndtri(torch.exp(torch.special.log_ndtr(-lower) + r.log()))

    # If sampled random number is very small, `ret` becomes inf.
    while torch.isinf(ret):
        r = torch.rand(torch.Size(()), dtype=torch.float64)
        ret = -torch.special.ndtri(torch.exp(torch.special.log_ndtr(-lower) + r.log()))

    return ret


_orthants_MVN_Gibbs_sampling_jit = torch.jit.script(_orthants_MVN_Gibbs_sampling)


def _compute_cov_diff_diff_inv(preferences: Tensor, cov_x_x: Tensor, noise_var: Tensor) -> Tensor:
    N = cov_x_x.shape[0]
    M = preferences.shape[0]

    # (sI + A K A^T)^-1 = s^-1 I - s^-2 A(K^-1 + s^-1 A^T A)^-1 A^T
    # (K^-1 + s^-1 A^T A)^-1 = K (I + s^-1 A^T A K)^-1  (To avoid computing K^-1)

    I_plus_sinv_AT_A_K = torch.eye(N, dtype=torch.float64)
    A_K = cov_x_x[preferences[:, 0], :] - cov_x_x[preferences[:, 1], :]
    I_plus_sinv_AT_A_K.index_add_(0, preferences[:, 0], A_K * (1 / noise_var))
    I_plus_sinv_AT_A_K.index_add_(0, preferences[:, 1], A_K * (-1 / noise_var))
    schur_inv: Tensor = torch.linalg.solve(I_plus_sinv_AT_A_K, cov_x_x, left=False)
    cov_diff_diff_inv = schur_inv[:, preferences[:, 0]] - schur_inv[:, preferences[:, 1]]
    cov_diff_diff_inv = (
        cov_diff_diff_inv[preferences[:, 0], :] - cov_diff_diff_inv[preferences[:, 1], :]
    )
    cov_diff_diff_inv *= -1 / noise_var**2
    idx_M = torch.arange(M)
    cov_diff_diff_inv[idx_M, idx_M] += 1.0 / noise_var

    return cov_diff_diff_inv


class _SampledGP(botorch.models.model.Model):
    def __init__(
        self,
        kernel_func: Callable[[Tensor, Tensor], Tensor],
        x: Tensor,
        preferences: Tensor,
        noise_var: Tensor,
        diff: Tensor,
    ) -> None:
        super().__init__()
        self.kernel_func = kernel_func
        self.x = x
        self.preferences = preferences
        self.diff = diff
        self.noise_var = noise_var
        self._cov_diff_diff_inv = _compute_cov_diff_diff_inv(
            preferences=preferences,
            cov_x_x=self.kernel_func(x, x),
            noise_var=noise_var,
        )

    def posterior(
        self,
        X: Tensor,
        output_indices: list[int] | None = None,
        observation_noise: bool = False,
        posterior_transform: Any | None = None,
        **kwargs: Any,
    ) -> botorch.posteriors.gpytorch.GPyTorchPosterior:
        assert posterior_transform is None
        assert output_indices is None
        assert self.x.shape[-1] == X.shape[-1]

        x_expanded = self.x.expand(X.shape[:-2] + (self.x.shape[-2], X.shape[-1]))

        cov_X_x = self.kernel_func(X, x_expanded)
        cov_X_diff = cov_X_x[..., self.preferences[:, 0]] - cov_X_x[..., self.preferences[:, 1]]

        mean = cov_X_diff @ (self._cov_diff_diff_inv @ self.diff)
        cov = self.kernel_func(X, X) - cov_X_diff @ self._cov_diff_diff_inv @ cov_X_diff.transpose(
            -1, -2
        )
        if observation_noise:
            idx = torch.arange(cov.shape[-1])
            cov[..., idx, idx] += self.noise_var

        return botorch.posteriors.gpytorch.GPyTorchPosterior(
            distribution=gpytorch.distributions.MultivariateNormal(
                mean=mean,
                covariance_matrix=cov,
            )
        )

    @property
    def batch_shape(self) -> torch.Size:
        return torch.Size()

    @property
    def num_outputs(self) -> int:
        return 1


def _truncnorm_mean_var_logz(alpha: Tensor) -> tuple[Tensor, Tensor, Tensor]:
    SQRT_HALF = math.sqrt(0.5)
    SQRT_HALF_PI = math.sqrt(0.5 * math.pi)
    logz = torch.special.log_ndtr(-alpha)
    mean = 1 / (SQRT_HALF_PI * torch.special.erfcx(alpha * SQRT_HALF))
    var = 1 - mean * (mean - alpha)
    return mean, var, logz


def _observation(var0: Tensor, mean0: Tensor, noise_var: Tensor) -> tuple[Tensor, Tensor, Tensor]:
    obs_var = var0 + noise_var
    obs_sigma = torch.sqrt(obs_var)
    alpha = -mean0 / torch.clamp_min(obs_sigma, min=1e-20)
    mean_norm, var_norm, logz = _truncnorm_mean_var_logz(alpha)

    denom_factor = 1 / torch.clamp_min(noise_var + var_norm * var0, min=1e-20)
    da = (1 - var_norm) * denom_factor
    db = (mean0 * (1 - var_norm) + obs_sigma * mean_norm) * denom_factor
    return (da, db, logz)


def _orthants_MVN_EP(
    cov0: Tensor, preferences: Tensor, noise_var: Tensor, cycles: int
) -> tuple[Tensor, Tensor, Tensor]:
    N = cov0.shape[0]
    M = preferences.shape[0]
    mu = torch.zeros(N, dtype=cov0.dtype)
    cov = cov0.clone()
    virtual_obs_a = [torch.tensor(0.0, dtype=cov0.dtype) for _ in range(M)]
    virtual_obs_b = [torch.tensor(0.0, dtype=cov0.dtype) for _ in range(M)]
    log_zs = torch.zeros(M, dtype=cov0.dtype)

    for _ in range(cycles):
        for i in range(M):
            pref_i = preferences[i, :]
            mean1 = mu[pref_i[0]] - mu[pref_i[1]]
            Sxy = cov[pref_i[0]] - cov[pref_i[1]]
            var1 = Sxy[pref_i[0]] - Sxy[pref_i[1]]

            r0 = (1 - var1 * virtual_obs_a[i]).reciprocal()
            var0 = var1 * r0
            mean0 = (mean1 - var1 * virtual_obs_b[i]) * r0

            virtual_obs_a2, virtual_obs_b2, logz = _observation(var0, mean0, noise_var)

            da = virtual_obs_a2 - virtual_obs_a[i]
            db = virtual_obs_b2 - virtual_obs_b[i]
            virtual_obs_a[i] = virtual_obs_a2
            virtual_obs_b[i] = virtual_obs_b2

            dr = (1 + var1 * da).reciprocal()
            mu = mu + Sxy * ((db - mean1 * da) * dr)
            cov = cov - (Sxy[:, None] * (da * dr)) @ Sxy[None, :]
            log_zs[i] = logz
    return mu, cov, torch.sum(log_zs)


_orthants_MVN_EP_jit = torch.jit.script(_orthants_MVN_EP)


class _PreferentialGP:
    def __init__(self, kernel: gpytorch.kernels.Kernel, noise_prior: Prior, dims: int) -> None:
        self.kernel = kernel
        self.noise_prior = noise_prior
        self.dims = dims

        self.diff = torch.empty((0,), dtype=torch.float64, requires_grad=False)
        self.log_noise = torch.nn.Parameter(
            torch.tensor(0.0, dtype=torch.float64), requires_grad=True
        )

    def fit_params_EP(self, X: Tensor, preferences: Tensor) -> None:
        if len(preferences) == 0:
            return
        tolerance = 1e-3
        max_iter = 100

        optim = torch.optim.LBFGS([*self.kernel.parameters(), self.log_noise])

        last_params = [p.detach().clone() for p in optim.param_groups[0]["params"]]
        for _ in range(max_iter):

            def closure() -> Tensor:
                optim.zero_grad()
                noise = self.log_noise.exp()
                cov0 = self.kernel.forward(X, X).to_dense()
                _, _, logz = _orthants_MVN_EP_jit(cov0, preferences, noise, cycles=2)

                loss = -logz - self.noise_prior.log_prob(noise)
                for _, _, prior, param, _ in self.kernel.named_priors():
                    loss = loss - prior.log_prob(param(self.kernel)).sum()

                loss.backward()
                return loss

            optim.step(closure)

            # Check for convergence
            params = optim.param_groups[0]["params"]
            for p_old, p_new in zip(last_params, params):
                if torch.max(torch.abs(p_old - p_new)) > tolerance:
                    break
            else:
                break
            last_params = [p.detach().clone() for p in params]

    def sample_gp(self, x: Tensor, preferences: Tensor) -> _SampledGP:
        self.fit_params_EP(x, preferences)

        with torch.no_grad():
            cov_diff_diff_inv = _compute_cov_diff_diff_inv(
                preferences=preferences,
                cov_x_x=self.kernel(x, x).to_dense(),
                noise_var=self.log_noise.exp(),
            )

            original_diff_size = len(self.diff)
            self.diff.resize_(len(preferences))
            self.diff[original_diff_size:] = 0.0

            self.diff = _orthants_MVN_Gibbs_sampling_jit(
                cov_inv=cov_diff_diff_inv,
                initial_sample=self.diff,
                cycles=20,
            )[-1]
            return _SampledGP(
                kernel_func=lambda x1, x2: self.kernel(x1, x2).to_dense(),
                x=x,
                preferences=preferences,
                noise_var=self.log_noise.exp(),
                diff=self.diff,
            )


class PreferentialGPSampler(optuna.samplers.BaseSampler):
    """Sampler for preferential optimization using Gaussian process.

    The sampling algorithm is based on `Takeno et al., 2023 <https://arxiv.org/abs/2302.01513>`_.
    This sampler uses BoTorch to optimize acquisition function.

    Args:
        kernel:
            Kernel that computes the covariance on the Gaussian process. Defaults to
            Matern 3/2 Kernel + ARD.
        noise_prior:
            Prior of the observation noise. Defaults to gamma prior.
        independent_sampler:
            A :class:`~optuna.samplers.BaseSampler` instance that is used for independent
            sampling. The parameters not contained in the relative search space are sampled
            by this sampler. If :obj:`None` is specified,
            :class:`~optuna.samplers.RandomSampler` is used as the default.
        seed:
            Seed for random number generator.
    """

    def __init__(
        self,
        *,
        kernel: gpytorch.kernels.Kernel | None = None,
        noise_prior: Prior | None = None,
        independent_sampler: optuna.samplers.BaseSampler | None = None,
        seed: int | None = None,
    ) -> None:
        self.kernel = kernel
        self.noise_prior = noise_prior or gpytorch.priors.GammaPrior(5.0, 50.0)

        self._rng = np.random.RandomState(seed)
        self.independent_sampler = independent_sampler or optuna.samplers.RandomSampler(
            seed=self._rng.randint(2**32, dtype=np.int64)  # type: ignore
        )

        self._search_space = optuna.search_space.IntersectionSearchSpace()
        self._gp: _PreferentialGP | None = None

    def reseed_rng(self) -> None:
        self.independent_sampler.reseed_rng()
        self._rng = np.random.RandomState()

    def infer_relative_search_space(
        self, study: optuna.Study, trial: optuna.trial.FrozenTrial
    ) -> dict[str, optuna.distributions.BaseDistribution]:
        return self._search_space.calculate(study)

    def sample_relative(
        self,
        study: optuna.Study,
        trial: optuna.trial.FrozenTrial,
        search_space: dict[str, optuna.distributions.BaseDistribution],
    ) -> dict[str, Any]:
        study_system_attrs = study._storage.get_study_system_attrs(study._study_id)
        preferences = get_preferences(study_system_attrs)
        if len(preferences) == 0 or len(search_space) == 0:
            return {}

        trials = study.get_trials(deepcopy=False)
        trials_with_preference = list({t for (b, w) in preferences for t in (b, w)})
        ids = {t: i for i, t in enumerate(trials_with_preference)}

        trans = optuna._transform._SearchSpaceTransform(
            search_space, transform_log=True, transform_step=True, transform_0_1=True
        )
        params = torch.tensor(
            np.array([trans.transform(trials[t].params) for t in trials_with_preference]),
            dtype=torch.float64,
        )
        pref_ids = torch.tensor([[ids[b], ids[w]] for b, w in preferences], dtype=torch.int32)
        with torch.random.fork_rng():
            torch.manual_seed(self._rng.randint(2**32, dtype=np.int64))

            self._gp = self._gp or _PreferentialGP(
                kernel=self.kernel
                or gpytorch.kernels.MaternKernel(
                    nu=1.5,
                    ard_num_dims=len(trans.bounds),
                    lengthscale_prior=gpytorch.priors.GammaPrior(5.0, 10.0),
                    lengthscale_constraint=gpytorch.constraints.GreaterThan(
                        0.0,
                        transform=torch.exp,
                        inv_transform=torch.log,
                    ),
                ),
                noise_prior=self.noise_prior,
                dims=len(trans.bounds),
            )
            if self._gp.dims != len(trans.bounds):
                raise NotImplementedError(
                    "The search space has changed. "
                    "Dynamic search space is not supported in PreferentialGPSampler."
                )

            sampled_gp = self._gp.sample_gp(params, pref_ids)
            acqf = botorch.acquisition.analytic.LogExpectedImprovement(
                model=sampled_gp,
                best_f=torch.max(sampled_gp.posterior(params[:, None, :]).mean),
            )

            # TODO: Make it possible to apply it on mixed search space
            def get_all_possible_params(dist: optuna.distributions.BaseDistribution) -> list[Any]:
                if isinstance(dist, CategoricalDistribution):
                    return list(dist.choices)
                elif (
                    isinstance(dist, (IntDistribution, FloatDistribution))
                    and dist.step is not None
                ):
                    return list(np.arange(dist.low, dist.high, dist.step))
                else:
                    return []

            all_possible_params = {
                name: get_all_possible_params(dist) for name, dist in search_space.items()
            }

            is_all_discrete = all(
                len(possible_params) > 0 for possible_params in all_possible_params.values()
            )
            search_space_size = math.prod(  # Don't use np.prod to avoid overflow
                [len(possible_params) for possible_params in all_possible_params.values()]
            )
            # TODO(contramundum53): Fix this arbitrarily chosen limit.
            size_limit = 1e6
            can_evaluate_all = is_all_discrete and search_space_size <= size_limit

            if (
                any(isinstance(dist, CategoricalDistribution) for dist in search_space.values())
                and not can_evaluate_all
            ):
                if is_all_discrete:
                    warnings.warn(
                        "The objective function has categorical parameters, "
                        "but the total search space is too large to be enumerated. "
                        f"(Search space size: {search_space_size} > limit: {size_limit})"
                        "This may result in significantly bad performance."
                    )
                else:
                    warnings.warn(
                        "The objective function has categorical parameters, "
                        "but the search space cannot be enumerated because "
                        "it also contains continuous parameters. "
                        "This may result in significantly bad performance. "
                        "You can work around this problem by specifying 'step' "
                        "in each continuous parameter."
                    )

            if is_all_discrete and can_evaluate_all:
                all_param_combinations = itertools.product(
                    *[
                        [(name, choice) for choice in possible_params]
                        for name, possible_params in all_possible_params.items()
                    ]
                )
                choices = torch.tensor(
                    np.array([trans.transform(dict(params)) for params in all_param_combinations]),
                    dtype=torch.float64,
                )
                candidates, _ = botorch.optim.optimize_acqf_discrete(
                    acq_function=acqf,
                    choices=choices,
                    q=1,
                )
            else:
                candidates, _ = botorch.optim.optimize_acqf(
                    acq_function=acqf,
                    bounds=torch.from_numpy(trans.bounds.T),
                    q=1,
                    num_restarts=10,
                    raw_samples=512,
                    options={"batch_limit": 5, "maxiter": 200},
                    sequential=True,
                )
            next_x = trans.untransform(candidates[0].detach().numpy())
            return next_x

    def sample_independent(
        self,
        study: optuna.Study,
        trial: optuna.trial.FrozenTrial,
        param_name: str,
        param_distribution: optuna.distributions.BaseDistribution,
    ) -> Any:
        warnings.warn(
            "Dynamic search space detected. "
            f"Falling back to {self.independent_sampler.__class__.__name__}."
        )

        return self.independent_sampler.sample_independent(
            study, trial, param_name, param_distribution
        )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING


_logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from typing import BinaryIO

    from optuna_dashboard.artifact.protocol import ArtifactBackend


class AppendPrefix:
    """An artifact backend middleware that appends a prefix string to artifact ids.

    Example:
       .. code-block:: python

          import optuna
          from optuna_dashboard.artifact import upload_artifact
          from optuna_dashboard.artifact.boto3 import Boto3Backend
          from optuna_dashboard.artifact.prefix import AppendPrefix

          artifact_backend = AppendPrefix(
             Boto3Backend("my-bucket"),
             prefix="my-folder/"
          )

          def objective(trial: optuna.Trial) -> float:
              ... = trial.suggest_float("x", -10, 10)
              file_path = generate_example_png(...)
              upload_artifact(artifact_backend, trial, file_path)
              return ...
    """

    def __init__(
        self,
        backend: ArtifactBackend,
        prefix: str,
    ) -> None:
        self._backend = backend
        self._prefix = prefix

    def _with_prefix(self, artifact_id: str) -> str:
        return self._prefix + artifact_id

    def open(self, artifact_id: str) -> BinaryIO:
        return self._backend.open(self._with_prefix(artifact_id))

    def write(self, artifact_id: str, content_body: BinaryIO) -> None:
        return self._backend.write(self._with_prefix(artifact_id), content_body)

    def remove(self, artifact_id: str) -> None:
        return self._backend.remove(self._with_prefix(artifact_id))


if TYPE_CHECKING:
    # A mypy-runtime assertion to ensure that SCSBackend
    # implements all abstract methods in ArtifactBackendProtocol.
    from optuna_dashboard.artifact.file_system import FileSystemBackend

    _: ArtifactBackend = AppendPrefix(FileSystemBackend("."), "prefix-")



from __future__ import annotations

import base64
import functools
import json
import logging
import traceback
from typing import Any
from typing import Callable
from typing import cast
from typing import Dict
from typing import TypeVar
from typing import Union

from bottle import BaseResponse
from bottle import HTTPError
from bottle import response


BottleViewReturn = Union[str, bytes, Dict[str, Any], BaseResponse]
BottleView = TypeVar("BottleView", bound=Callable[..., BottleViewReturn])
BottleAPIView = TypeVar("BottleAPIView", bound=Callable[..., Dict[str, Any]])
logger = logging.getLogger(__name__)


def json_api_view(view: BottleAPIView) -> BottleAPIView:
    @functools.wraps(view)
    def decorated(*args: list[Any], **kwargs: dict[str, Any]) -> BottleViewReturn:
        try:
            response.content_type = "application/json"
            response_body = view(*args, **kwargs)
            return response_body
        except HTTPError as e:
            response.status = e.status_code
            return json.dumps({"reason": str(e.body)})
        except Exception as e:
            response.status = 500
            response.content_type = "application/json"
            stacktrace = "\n".join(traceback.format_tb(e.__traceback__))
            logger.error(f"Exception: {e}\n{stacktrace}")
            return json.dumps({"reason": "internal server error"})

    return cast(BottleAPIView, decorated)


def parse_data_uri(data_uri: str) -> tuple[str, bytes]:
    prefix, a = data_uri.split(":", 1)
    if prefix != "data":
        raise ValueError("data url must start with 'data:' prefix")
    mediatype_with_suffix, base64_data = a.split(",", 1)
    mediatype = mediatype_with_suffix.split(";", 1)[0]
    data = base64.standard_b64decode(base64_data)
    return mediatype, data



from __future__ import annotations

from typing import Any
from typing import Optional
import warnings

import optuna


# Should be equivalent to `optuna.study.study._SYSTEM_ATTR_METRIC_NAMES`.
# See https://github.com/optuna/optuna/pull/4383 for details.
SYSTEM_ATTR_METRIC_NAMES = "study:metric_names"

SYSTEM_ATTR_NAME = "dashboard:objective_names"


def set_objective_names(study: optuna.Study, names: list[str]) -> None:
    """Set the names of objectives.

    Example:

       .. code-block:: python

          import optuna
          from optuna_dashboard import set_objective_names

          study = optuna.create_study(directions=["minimize", "minimize"])
          set_objective_names(study, ["val_loss", "flops"])
    """

    if hasattr(study, "set_metric_names"):
        warnings.warn(
            "`set_objective_names()` function is deprecated."
            " Please use `study.set_metric_names()` instead."
            " See https://optuna-dashboard.readthedocs.io/en/latest/errors.html for details.",
            category=FutureWarning,
        )
        study.set_metric_names(names)
        return

    storage = study._storage
    study_id = study._study_id

    directions = storage.get_study_directions(study_id)
    if len(directions) != len(names):
        raise ValueError("names must be the same length with the number of objectives.")
    storage.set_study_system_attr(study_id, SYSTEM_ATTR_NAME, names)


def get_objective_names(system_attrs: dict[str, Any]) -> Optional[list[str]]:
    if SYSTEM_ATTR_METRIC_NAMES in system_attrs:
        return system_attrs[SYSTEM_ATTR_METRIC_NAMES]
    return system_attrs.get(SYSTEM_ATTR_NAME)



from __future__ import annotations

from typing import TYPE_CHECKING


try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore


if TYPE_CHECKING:
    from typing import BinaryIO


class ArtifactBackend(Protocol):
    """A protocol defining the interface for an artifact backend.

    An artifact backend is responsible for managing the storage and retrieval
    of artifact data. The backend should provide methods for opening, writing
    and removing artifacts.
    """

    def open(self, artifact_id: str) -> BinaryIO:
        """Open the artifact identified by the artifact_id.

        This method should return a binary file-like object in read mode, similar to
        ``open(..., mode="rb")``. If the artifact does not exist, an
        :exc:`~optuna_dashboard.artifact.exceptions.ArtifactNotFound` exception
        should be raised.

        Args:
            artifact_id: The identifier of the artifact to open.

        Returns:
            BinaryIO: A binary file-like object that can be read from.
        """
        ...

    def write(self, artifact_id: str, content_body: BinaryIO) -> None:
        """Save the content to the backend.

        Args:
            artifact_id: The identifier of the artifact to write to.
            content_body: The content to write to the artifact.
        """
        ...

    def remove(self, artifact_id: str) -> None:
        """Remove the artifact identified by the artifact_id.

        This method should delete the artifact from the backend. If the artifact does not
        exist, an :exc:`~optuna_dashboard.artifact.exceptions.ArtifactNotFound` exception
        may be raised.

        Args:
            artifact_id: The identifier of the artifact to remove.
        """
        ...



from __future__ import annotations

import logging
import time
from typing import TYPE_CHECKING

from optuna_dashboard.artifact.exceptions import ArtifactNotFound


_logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from typing import BinaryIO

    from optuna_dashboard.artifact.protocol import ArtifactBackend


class Backoff:
    """An artifact backend middleware for exponential backoff.

    Example:
       .. code-block:: python

          import optuna
          from optuna_dashboard.artifact import upload_artifact
          from optuna_dashboard.artifact.backoff import Backoff
          from optuna_dashboard.artifact.boto3 import Boto3Backend

          artifact_backend = Backoff(Boto3Backend("my-bucket"))

          def objective(trial: optuna.Trial) -> float:
              ... = trial.suggest_float("x", -10, 10)
              file_path = generate_example_png(...)
              upload_artifact(artifact_backend, trial, file_path)
              return ...
    """

    def __init__(
        self,
        backend: ArtifactBackend,
        max_retries: int = 10,
        multiplier: float = 2,
        min_delay: float = 0.1,
        max_delay: float = 30,
    ) -> None:
        # Default sleep seconds:
        # 0.1, 0.2, 0.4, 0.8, 1.6, 3.2, 6.4, 12.8, 25.6, 30
        self._backend = backend
        assert max_retries > 0
        assert multiplier > 0
        assert min_delay > 0
        assert max_delay > min_delay
        self._max_retries = max_retries
        self._multiplier = multiplier
        self._min_delay = min_delay
        self._max_delay = max_delay

    def _get_sleep_secs(self, n_retry: int) -> float:
        return min(self._min_delay * self._multiplier**n_retry, self._max_delay)

    def open(self, artifact_id: str) -> BinaryIO:
        for i in range(self._max_retries):
            try:
                return self._backend.open(artifact_id)
            except ArtifactNotFound:
                raise
            except Exception as e:
                if i == self._max_retries - 1:
                    raise
                else:
                    _logger.error(f"Failed to open artifact={artifact_id} n_retry={i}", exc_info=e)
            time.sleep(self._get_sleep_secs(i))
        assert False, "must not reach here"

    def write(self, artifact_id: str, content_body: BinaryIO) -> None:
        for i in range(self._max_retries):
            try:
                self._backend.write(artifact_id, content_body)
                break
            except ArtifactNotFound:
                raise
            except Exception as e:
                if i == self._max_retries - 1:
                    raise
                else:
                    _logger.error(f"Failed to open artifact={artifact_id} n_retry={i}", exc_info=e)
            content_body.seek(0)
            time.sleep(self._get_sleep_secs(i))

    def remove(self, artifact_id: str) -> None:
        for i in range(self._max_retries):
            try:
                self._backend.remove(artifact_id)
            except ArtifactNotFound:
                raise
            except Exception as e:
                if i == self._max_retries - 1:
                    raise
                else:
                    _logger.error(f"Failed to delete artifact={artifact_id}", exc_info=e)
            time.sleep(self._get_sleep_secs(i))


if TYPE_CHECKING:
    # A mypy-runtime assertion to ensure that SCSBackend
    # implements all abstract methods in ArtifactBackendProtocol.
    from optuna_dashboard.artifact.file_system import FileSystemBackend

    _: ArtifactBackend = Backoff(FileSystemBackend("."))



from __future__ import annotations

import os
import shutil
from typing import TYPE_CHECKING
import warnings

from optuna_dashboard.artifact.exceptions import ArtifactNotFound


if TYPE_CHECKING:
    from typing import BinaryIO


class FileSystemBackend:
    """An artifact backend for file systems.

    .. warning::

       This class is deprecated. Please use `optuna.artifacts.FileSystemArtifactStore
       <https://optuna.readthedocs.io/en/latest/reference/generated/optuna.artifacts.
       FileSystemArtifactStore.html>`_ instead.

    Example:
       .. code-block:: python

          import optuna
          from optuna_dashboard.artifact import upload_artifact
          from optuna_dashboard.artifact.file_system import FileSystemBackend

          artifact_backend = FileSystemBackend("./artifacts")

          def objective(trial: optuna.Trial) -> float:
              ... = trial.suggest_float("x", -10, 10)
              file_path = generate_example_png(...)
              upload_artifact(artifact_backend, trial, file_path)
              return ...
    """

    def __init__(self, base_path: str) -> None:
        self._base_path = base_path
        warnings.warn(
            "FileSystemBackend is deprecated. Please use FileSystemArtifactStore instead.\n"
            "See https://optuna-dashboard.readthedocs.io/en/latest/errors.html for details",
            DeprecationWarning,
        )

    def open(self, artifact_id: str) -> BinaryIO:
        filepath = os.path.join(self._base_path, artifact_id)
        try:
            f = open(filepath, "rb")
        except FileNotFoundError as e:
            raise ArtifactNotFound("not found") from e
        return f

    def write(self, artifact_id: str, content_body: BinaryIO) -> None:
        filepath = os.path.join(self._base_path, artifact_id)
        with open(filepath, "wb") as f:
            shutil.copyfileobj(content_body, f)

    def remove(self, artifact_id: str) -> None:
        filepath = os.path.join(self._base_path, artifact_id)
        try:
            os.remove(filepath)
        except FileNotFoundError as e:
            raise ArtifactNotFound("not found") from e


if TYPE_CHECKING:
    # A mypy-runtime assertion to ensure that LocalArtifactBackend
    # implements all abstract methods in ArtifactBackendProtocol.
    from .protocol import ArtifactBackend

    _: ArtifactBackend = FileSystemBackend("")



from __future__ import annotations

import math
from typing import TYPE_CHECKING
import uuid

from optuna import Study


if TYPE_CHECKING:
    from typing import Any

    from optuna.storages import BaseStorage
    import plotly.graph_objs as go


SYSTEM_ATTR_PLOT_DATA = "dashboard:plot_data:"
SYSTEM_ATTR_MAX_LENGTH = 2045


def save_plotly_graph_object(
    study: Study, figure: go.Figure, *, graph_object_id: str | None = None
) -> str:
    """Save the user-defined plotly's graph object to the study.

    Example:

       .. code-block:: python

          import optuna
          from optuna_dashboard import save_plotly_graph_object

          def objective(trial):
              x = trial.suggest_float("x", -100, 100)
              y = trial.suggest_categorical("y", [-1, 0, 1])
              return x**2 + y

          study = optuna.create_study()
          study.optimize(objective, n_trials=100)

          figure = optuna.visualization.plot_optimization_history(study)
          save_plotly_graph_object(study, figure)

    Args:
        study:
            Target study object.
        figure:
            A :class:`plotly.graph_objects.Figure` object to save.
        graph_object_id:
            Unique identifier of the graph object. If specified, the graph object is overwritten.
            This must be a valid HTML id attribute value.

    Returns:
        The graph object ID.
    """
    if graph_object_id is not None and not is_valid_graph_object_id(graph_object_id):
        raise ValueError("graph_object_id must be a valid HTML id attribute value.")

    storage = study._storage
    study_id = study._study_id

    graph_object_id = graph_object_id or str(uuid.uuid4())
    key = SYSTEM_ATTR_PLOT_DATA + graph_object_id + ":"
    plot_data_json_str = figure.to_json()
    save_graph_object_json(storage, study_id, key, plot_data_json_str)
    return graph_object_id


def save_graph_object_json(
    storage: BaseStorage, study_id: int, key_prefix: str, plot_data_json_str: str
) -> None:
    plot_data_system_attrs = split_plot_data(plot_data_json_str, key_prefix)
    for k, v in plot_data_system_attrs.items():
        storage.set_study_system_attr(study_id, k, v)

    # Clear previous graph object attributes
    study_system_attrs = storage.get_study_system_attrs(study_id)
    all_plot_data_system_attrs = [k for k in study_system_attrs if k.startswith(key_prefix)]
    if len(all_plot_data_system_attrs) > len(plot_data_system_attrs):
        for i in range(len(plot_data_system_attrs), len(all_plot_data_system_attrs)):
            storage.set_study_system_attr(study_id, f"{key_prefix}{i}", "")


def list_graph_object_ids(system_attrs: dict[str, Any]) -> list[str]:
    titles = set()
    for key in system_attrs:
        if not key.startswith(SYSTEM_ATTR_PLOT_DATA):
            continue

        s = key.split(":", maxsplit=2)  # e.g. ["dashboard", "plot_data", "Optimization History:1"]
        if len(s) != 3:
            continue
        # Please note that title may contain ":".
        title = s[2].rsplit(":", maxsplit=1)[0]
        titles.add(title)
    return list(titles)


def get_plotly_graph_objects(system_attrs: dict[str, Any]) -> dict[str, str]:
    graph_objects = {}
    for title in list_graph_object_ids(system_attrs):
        key_prefix = SYSTEM_ATTR_PLOT_DATA + title + ":"
        plot_data_attrs = {k: v for k, v in system_attrs.items() if k.startswith(key_prefix)}
        graph_objects[title] = concat_plot_data(plot_data_attrs, key_prefix)
    return graph_objects


def split_plot_data(plot_data_str: str, key_prefix: str) -> dict[str, str]:
    plot_data_len = len(plot_data_str)
    attrs = {}
    for i in range(math.ceil(plot_data_len / SYSTEM_ATTR_MAX_LENGTH)):
        start = i * SYSTEM_ATTR_MAX_LENGTH
        end = min((i + 1) * SYSTEM_ATTR_MAX_LENGTH, plot_data_len)
        attrs[f"{key_prefix}{i}"] = plot_data_str[start:end]
    return attrs


def concat_plot_data(plot_data_attrs: dict[str, str], key_prefix: str) -> str:
    return "".join(plot_data_attrs[f"{key_prefix}{i}"] for i in range(len(plot_data_attrs)))


def is_valid_graph_object_id(graph_object_id: str) -> bool:
    if len(graph_object_id) == 0:
        return False

    # Can only contain letters [A-Za-z], numbers [0-9], hyphens ("-"), underscores ("_"),
    # colons, and periods.
    if not all(
        "a" <= c <= "z" or "A" <= c <= "Z" or "0" <= c <= "9" or c in ("-", "_", ":", ".")
        for c in graph_object_id[1:]
    ):
        return False
    # Unlike HTML id attribute, graph object id can begin with a letter [A-Za-z]
    return True



from __future__ import annotations

import argparse
import os
from socketserver import ThreadingMixIn
import sys
from typing import TYPE_CHECKING
from wsgiref.simple_server import make_server
from wsgiref.simple_server import WSGIServer

from bottle import Bottle
from bottle import run
from optuna.storages import BaseStorage
from optuna.storages import RDBStorage
from optuna.version import __version__ as optuna_ver
from packaging import version

from . import __version__
from ._app import create_app
from ._sql_profiler import register_profiler_view
from ._storage_url import get_storage
from .artifact._backend_to_store import ArtifactBackendToStore
from .artifact.file_system import FileSystemBackend


if TYPE_CHECKING:
    from typing import Literal

    from optuna.artifacts._protocol import ArtifactStore


DEBUG = os.environ.get("OPTUNA_DASHBOARD_DEBUG") == "1"
SERVER_CHOICES = ["auto", "wsgiref", "gunicorn"]


class ThreadedWSGIServer(ThreadingMixIn, WSGIServer):
    pass


def run_wsgiref(app: Bottle, host: str, port: int, quiet: bool) -> None:
    print(f"Listening on http://{host}:{port}/", file=sys.stderr)
    print("Hit Ctrl-C to quit.\n", file=sys.stderr)
    httpd = make_server(host, port, app, server_class=ThreadedWSGIServer)
    httpd.serve_forever()


def run_gunicorn(app: Bottle, host: str, port: int, quiet: bool) -> None:
    # See https://docs.gunicorn.org/en/latest/custom.html

    from gunicorn.app.base import BaseApplication

    class Application(BaseApplication):
        def load_config(self) -> None:
            self.cfg.set("bind", f"{host}:{port}")
            self.cfg.set("threads", 4)
            if quiet:
                self.cfg.set("loglevel", "error")

        def load(self) -> Bottle:
            return app

    Application().run()


def run_debug_server(app: Bottle, host: str, port: int, quiet: bool) -> None:
    run(
        app,
        host=host,
        port=port,
        server="wsgiref",
        quiet=quiet,
        reloader=DEBUG,
    )


def auto_select_server(
    server_arg: Literal["auto", "gunicorn", "wsgiref"],
) -> Literal["gunicorn", "wsgiref"]:
    if server_arg != "auto":
        return server_arg

    try:
        import gunicorn  # NOQA

        return "gunicorn"
    except ImportError:
        return "wsgiref"


def main() -> None:
    parser = argparse.ArgumentParser(description="Real-time dashboard for Optuna.")
    parser.add_argument("storage", help="Storage URL (e.g. sqlite:///example.db)", type=str)
    parser.add_argument(
        "--storage-class",
        help="Storage class hint (e.g. JournalFileStorage)",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--port", help="port number (default: %(default)s)", type=int, default=8080
    )
    parser.add_argument("--host", help="hostname (default: %(default)s)", default="127.0.0.1")
    parser.add_argument(
        "--server",
        help="server (default: %(default)s)",
        default="auto",
        choices=SERVER_CHOICES,
    )
    parser.add_argument(
        "--artifact-dir",
        help="directory to store artifact files",
        default=None,
    )
    parser.add_argument("--version", "-v", action="version", version=__version__)
    parser.add_argument("--quiet", "-q", help="quiet", action="store_true")
    args = parser.parse_args()

    storage: BaseStorage
    storage = get_storage(args.storage, storage_class=args.storage_class)

    artifact_store: ArtifactStore | None
    if args.artifact_dir is None:
        artifact_store = None
    elif version.parse(optuna_ver) >= version.Version("3.3.0"):
        from optuna.artifacts import FileSystemArtifactStore

        artifact_store = FileSystemArtifactStore(args.artifact_dir)
    else:
        artifact_backend = FileSystemBackend(args.artifact_dir)
        artifact_store = ArtifactBackendToStore(artifact_backend)
    app = create_app(storage, artifact_store=artifact_store, debug=DEBUG)

    if DEBUG and isinstance(storage, RDBStorage):
        app = register_profiler_view(app, storage)

    server = auto_select_server(args.server)
    if DEBUG:
        run_debug_server(app, args.host, args.port, args.quiet)
    elif server == "wsgiref":
        run_wsgiref(app, args.host, args.port, args.quiet)
    elif server == "gunicorn":
        run_gunicorn(app, args.host, args.port, args.quiet)
    else:
        raise Exception("must not reach here")


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import TYPE_CHECKING

import optuna
from optuna.trial import FrozenTrial

import streamlit as st

from .._form_widget import get_form_widgets_json
from .._note import get_note_from_system_attrs


if TYPE_CHECKING:
    from typing import Callable
    from typing import Optional
    from typing import Sequence
    from typing import Union

    from .._form_widget import ChoiceWidgetJSON
    from .._form_widget import SliderWidgetJSON
    from .._form_widget import TextInputWidgetJSON
    from .._form_widget import UserAttrRefJSON


def render_trial_note(study: optuna.Study, trial: FrozenTrial) -> None:
    """Write a trial note to UI with streamlit as a markdown format.

    Args:
        study: The optuna study object.
        trial: The optuna trial object to get note.
    """

    note = get_note_from_system_attrs(study.system_attrs, trial._trial_id)
    st.markdown(note["body"], unsafe_allow_html=True)


def _format_choice(choice: float, widget: ChoiceWidgetJSON) -> str:
    return widget["choices"][widget["values"].index(choice)]


def _format_description(description: Optional[str]) -> str:
    return "" if description is None else description


def _render_widgets(
    widgets: Sequence[
        Union[ChoiceWidgetJSON, SliderWidgetJSON, TextInputWidgetJSON, UserAttrRefJSON]
    ],
    trial: FrozenTrial,
) -> tuple[bool, list[Optional[Union[str, float]]]]:
    values: list[Optional[Union[str, float]]] = []

    with st.form("user_input", clear_on_submit=False):
        for i, widget in enumerate(widgets):
            if widget["type"] == "choice":
                value = st.radio(
                    _format_description(widget["description"]),
                    widget["values"],
                    format_func=lambda choice, widget=widget: _format_choice(  # type: ignore
                        choice, widget
                    ),
                    horizontal=True,
                    key=f"radio_{i}",
                )
            elif widget["type"] == "slider":
                # NOTE: It is difficult to reflect "labels".
                value = st.slider(
                    _format_description(widget["description"]),
                    min_value=widget["min"],
                    max_value=widget["max"],
                    step=widget["step"],
                    key=f"slider_{i}",
                )
            elif widget["type"] == "text":
                # NOTE: Current implementation ignores "optional".
                value = st.text_input(_format_description(widget["description"]), key=f"text_{i}")  # type: ignore
            elif widget["type"] == "user_attr":
                value = trial.user_attrs[widget["key"]]
            else:
                raise ValueError(
                    "Widget type should be 'choice', 'slider', 'text', or 'user_attr'."
                )
            values.append(value)

        submitted = st.form_submit_button("Submit")
    return submitted, values


def render_user_attr_form_widgets(
    study: optuna.Study,
    trial: FrozenTrial,
    on_success_callback: Optional[Callable[[], None]] = None,
) -> None:
    """Render user input widgets to UI with streamlit.

    Submitted values to the forms are registered as each trial's user_attrs.

    Args:
        study: The optuna study object to get widget specification.
        trial: The optuna trial object to save user feedbacks.
        on_success_callback: The callback function which will be executed
                             when feedback submission is succeeded.

    Raises:
        ValueError: If No form widgets registered.
        ValueError: If 'output_type' of form widgets is not 'user_attr'.
    """

    form_widgets_dict = get_form_widgets_json(study.system_attrs)
    if form_widgets_dict is None:
        raise ValueError("No form widgets registered.")
    if form_widgets_dict["output_type"] != "user_attr":
        raise ValueError("'output_type' should be 'user_attr'.")

    widgets = form_widgets_dict["widgets"]
    submitted, values = _render_widgets(widgets, trial)

    if submitted:
        for widget, value in zip(widgets, values):
            if "user_attr_key" in widget.keys():
                study._storage.set_trial_user_attr(
                    trial._trial_id,
                    key=widget["user_attr_key"],  # type: ignore
                    value=value,
                )

        if on_success_callback is None:
            st.success("Submitted!")
        else:
            on_success_callback()


def render_objective_form_widgets(
    study: optuna.Study,
    trial: FrozenTrial,
    on_success_callback: Optional[Callable[[], None]] = None,
) -> None:
    """Render user input widgets to UI with streamlit.

    Submitted values to the forms are telled to optuna trial object.
    All submitted values should be float.
    Multiple widgets correspond to multi-objective optimization.

    Args:
        study: The optuna study object to get widget specification.
        trial: The optuna trial object to tell user feedbacks.
        on_success_callback: The callback function which will be executed
                             when feedback submission is succeeded.

    Raises:
        ValueError: If No form widgets registered.
        ValueError: If 'output_type' of form widgets is not 'objective'.
        ValueError: If any submitted values cannot be converted to float.
    """

    form_widgets_dict = get_form_widgets_json(study.system_attrs)
    if form_widgets_dict is None:
        raise ValueError("No form widgets registered.")
    if form_widgets_dict["output_type"] != "objective":
        raise ValueError("'output_type' should be 'objective'.")

    submitted, values = _render_widgets(form_widgets_dict["widgets"], trial)

    if submitted:
        values_float = []
        try:
            for value in values:
                values_float.append(float(value))  # type: ignore

            study.tell(trial.number, values_float)

            if on_success_callback is None:
                st.success("Submitted!")
            else:
                on_success_callback()
        except ValueError:
            st.error("Please enter float values.")



from __future__ import annotations

from typing import Any
from typing import TYPE_CHECKING

from optuna.storages import BaseStorage

from .preferential._study import PreferentialStudy


if TYPE_CHECKING:
    from typing import Literal

    OUTPUT_COMPONENT_TYPE = Literal["note", "artifact"]

_SYSTEM_ATTR_FEEDBACK_COMPONENT = "preference:component"


def _register_preference_feedback_component(
    study_id: int,
    storage: BaseStorage,
    component_type: OUTPUT_COMPONENT_TYPE,
    artifact_key: str | None = None,
) -> None:
    value: dict[str, Any] = {"output_type": component_type}
    if artifact_key is not None:
        value["artifact_key"] = artifact_key
    storage.set_study_system_attr(
        study_id=study_id,
        key=_SYSTEM_ATTR_FEEDBACK_COMPONENT,
        value=value,
    )


def register_preference_feedback_component(
    study: PreferentialStudy,
    component_type: OUTPUT_COMPONENT_TYPE,
    artifact_key: str | None = None,
) -> None:
    """Register a preference feedback component to the study.

    With this feature, you can change the component, displayed on the
    human feedback pages. By default, the Markdown note (``component_type="note"``)
    is displayed.  If you specify ``component_type="artifact"``, the viewer for the
    specified artifact file will be displayed.

    Args:
        study:
            The study to register the preference feedback component.
        component_type:
            The component type, displayed on the human feedback pages
            (default: ``"note"``).
        user_attr_artifact_key:
            This option is required when the ``component_type`` is ``"artifact"``.
            The user attribute, which is specified this field, must contain the
            ``artifact``id you want to display on the human feedback page.
    """
    if component_type == "artifact":
        assert (
            artifact_key is not None
        ), "artifact_key must be specified when component_type is Artifact"

    _register_preference_feedback_component(
        study_id=study._study._study_id,
        storage=study._study._storage,
        component_type=component_type,
        artifact_key=artifact_key,
    )



from __future__ import annotations

import io
import shutil
from typing import TYPE_CHECKING
import warnings

import boto3
from botocore.exceptions import ClientError
from optuna_dashboard.artifact.exceptions import ArtifactNotFound


if TYPE_CHECKING:
    from typing import BinaryIO

    from mypy_boto3_s3 import S3Client


class Boto3Backend:
    """An artifact backend for S3.

    .. warning::

       This class is deprecated. Please use `optuna.artifacts.Boto3ArtifactStore
       <https://optuna.readthedocs.io/en/latest/reference/generated/optuna.artifacts.
       Boto3ArtifactStore.html>`_ instead.

    Example:
       .. code-block:: python

          import optuna
          from optuna_dashboard.artifact import upload_artifact
          from optuna_dashboard.artifact.boto3 import Boto3Backend

          artifact_backend = Boto3Backend("my-bucket")

          def objective(trial: optuna.Trial) -> float:
              ... = trial.suggest_float("x", -10, 10)
              file_path = generate_example_png(...)
              upload_artifact(artifact_backend, trial, file_path)
              return ...
    """

    def __init__(
        self, bucket_name: str, client: S3Client | None = None, *, avoid_buf_copy: bool = False
    ) -> None:
        self.bucket = bucket_name
        self.client = client or boto3.client("s3")
        # This flag is added to avoid that upload_fileobj() method of Boto3 client
        # may close the source file object.
        # See https://github.com/boto/boto3/issues/929
        self._avoid_buf_copy = avoid_buf_copy
        warnings.warn(
            "Boto3Backend is deprecated. Please use Boto3ArtifactStore instead.\n"
            "See https://optuna-dashboard.readthedocs.io/en/latest/errors.html for details",
            DeprecationWarning,
        )

    def open(self, artifact_id: str) -> BinaryIO:
        try:
            obj = self.client.get_object(Bucket=self.bucket, Key=artifact_id)
        except ClientError as e:
            if _is_not_found_error(e):
                raise ArtifactNotFound("not found") from e
            raise
        body = obj.get("Body")
        assert body is not None
        return body  # type: ignore

    def write(self, artifact_id: str, content_body: BinaryIO) -> None:
        fsrc: BinaryIO = content_body
        if not self._avoid_buf_copy:
            buf = io.BytesIO()
            shutil.copyfileobj(content_body, buf)
            buf.seek(0)
            fsrc = buf
        self.client.upload_fileobj(fsrc, self.bucket, artifact_id)

    def remove(self, artifact_id: str) -> None:
        try:
            self.client.delete_object(Bucket=self.bucket, Key=artifact_id)
        except ClientError as e:
            if _is_not_found_error(e):
                raise ArtifactNotFound("not found") from e
            raise


def _is_not_found_error(e: ClientError) -> bool:
    error_code = e.response.get("Error", {}).get("Code")
    http_status_code = e.response.get("ResponseMetadata", {}).get("HTTPStatusCode")
    return error_code == "NoSuchKey" or http_status_code == 404


if TYPE_CHECKING:
    # A mypy-runtime assertion to ensure that Boto3Backend
    # implements all abstract methods in ArtifactBackendProtocol.
    from .protocol import ArtifactBackend

    _: ArtifactBackend = Boto3Backend("")



from __future__ import annotations

import math
from typing import TYPE_CHECKING

import optuna
from optuna.storages import BaseStorage


if TYPE_CHECKING:
    from typing import Any
    from typing import Optional
    from typing import TypedDict

    NoteType = TypedDict(
        "NoteType",
        {
            "version": int,
            "body": str,
        },
    )

SYSTEM_ATTR_MAX_LENGTH = 2045


def save_note(study_or_trial: optuna.Study | optuna.Trial, body: str) -> None:
    """Save the note (Markdown format) to the Study or Trial.

    Example:

       .. code-block:: python

          import optuna
          from optuna_dashboard import save_note


          def objective(trial: optuna.Trial) -> float:
              x1 = trial.suggest_float("x1", 0, 10)

              save_note(trial, textwrap.dedent(f'''\
              ## Trial {trial.number}

              You can *freely* take a **note** that is associated with the Trial.
              '''))
              return (x1 - 2) ** 2


          study = optuna.create_study()
          save_note(study, textwrap.dedent(f'''\
          ## {study.study_name}

          You can *freely* take a **note** that is associated with the study.
          '''))
          study.optimize(objective, n_trials=10)
    """
    trial_id: Optional[int] = None
    if isinstance(study_or_trial, optuna.Study):
        storage = study_or_trial._storage
        study_id = study_or_trial._study_id
    else:
        storage = study_or_trial.storage
        study_id = study_or_trial.study._study_id
        trial_id = study_or_trial._trial_id

    system_attrs = storage.get_study_system_attrs(study_id)
    next_ver = system_attrs.get(note_ver_key(trial_id), 0) + 1
    save_note_with_version(storage, study_id, trial_id, next_ver, body)


def get_note(study_or_trial: optuna.Study | optuna.Trial) -> str:
    """Get the note (Markdown format) from the Study or Trial.

    Example:

       .. code-block:: python

          import optuna
          from optuna_dashboard import save_note, get_note

          study = optuna.create_study()
          save_note(study, "**Hello** World")

          text = get_note(study)
          print(text)  # '**Hello** World'
    """
    trial_id: Optional[int] = None
    if isinstance(study_or_trial, optuna.Study):
        storage = study_or_trial._storage
        study_id = study_or_trial._study_id
    else:
        storage = study_or_trial.storage
        study_id = study_or_trial.study._study_id
        trial_id = study_or_trial._trial_id
    system_attrs = storage.get_study_system_attrs(study_id)
    note = get_note_from_system_attrs(system_attrs, trial_id)
    return note["body"]


def note_ver_key(trial_id: Optional[int]) -> str:
    prefix = "dashboard:note_ver"
    if trial_id is None:
        return prefix
    return f"dashboard:{trial_id}:note_ver"


def note_str_key_prefix(trial_id: Optional[int]) -> str:
    prefix = "dashboard:note_str:"
    if trial_id is None:
        return prefix
    return f"dashboard:{trial_id}:note_str:"


def copy_notes(storage: BaseStorage, src_study: optuna.Study, dst_study: optuna.Study) -> None:
    system_attrs = storage.get_study_system_attrs(study_id=src_study._study_id)

    # Copy individual trial notes
    for src_trial, dst_trial in zip(src_study.get_trials(), dst_study.get_trials()):
        note = get_note_from_system_attrs(system_attrs, src_trial._trial_id)["body"]
        save_note_with_version(storage, dst_study._study_id, dst_trial._trial_id, 0, note)

    # Copy study note
    note = get_note_from_system_attrs(system_attrs, None)["body"]
    save_note_with_version(storage, dst_study._study_id, None, 0, note)


def get_note_from_system_attrs(system_attrs: dict[str, Any], trial_id: Optional[int]) -> NoteType:
    if note_ver_key(trial_id) not in system_attrs:
        return {
            "version": 0,
            "body": "",
        }
    note_ver = int(system_attrs[note_ver_key(trial_id)])
    note_attrs: dict[str, str] = {
        key: value
        for key, value in system_attrs.items()
        if key.startswith(note_str_key_prefix(trial_id))
    }
    return {"version": note_ver, "body": concat_body(note_attrs, trial_id)}


def version_is_incremented(
    system_attrs: dict[str, Any], trial_id: Optional[int], req_note_ver: int
) -> bool:
    db_note_ver = system_attrs.get(note_ver_key(trial_id), 0)
    return req_note_ver == db_note_ver + 1


def save_note_with_version(
    storage: BaseStorage, study_id: int, trial_id: Optional[int], ver: int, body: str
) -> None:
    storage.set_study_system_attr(study_id, note_ver_key(trial_id), ver)

    attrs = split_body(body, trial_id)
    for k, v in attrs.items():
        storage.set_study_system_attr(study_id, k, v)

    # Clear previous messages
    all_note_attrs: dict[str, str] = {
        key: value
        for key, value in storage.get_study_system_attrs(study_id).items()
        if key.startswith(note_str_key_prefix(trial_id))
    }
    if len(all_note_attrs) > len(attrs):
        for i in range(len(attrs), len(all_note_attrs)):
            storage.set_study_system_attr(study_id, f"{note_str_key_prefix(trial_id)}{i}", "")


def split_body(note_str: str, trial_id: Optional[int]) -> dict[str, str]:
    note_len = len(note_str)
    attrs = {}
    for i in range(math.ceil(note_len / SYSTEM_ATTR_MAX_LENGTH)):
        start = i * SYSTEM_ATTR_MAX_LENGTH
        end = min((i + 1) * SYSTEM_ATTR_MAX_LENGTH, note_len)
        attrs[f"{note_str_key_prefix(trial_id)}{i}"] = note_str[start:end]
    return attrs


def concat_body(note_attrs: dict[str, str], trial_id: Optional[int]) -> str:
    return "".join(
        note_attrs[f"{note_str_key_prefix(trial_id)}{i}"] for i in range(len(note_attrs))
    )



from __future__ import annotations

from dataclasses import dataclass
import json
from typing import TYPE_CHECKING
from typing import Union
import warnings

import optuna
from optuna.version import __version__ as optuna_ver
from packaging import version


if TYPE_CHECKING:
    from typing import Any
    from typing import Literal
    from typing import Optional
    from typing import Sequence
    from typing import TypedDict

    ChoiceWidgetJSON = TypedDict(
        "ChoiceWidgetJSON",
        {
            "type": Literal["choice"],
            "description": Optional[str],
            "choices": list[str],
            "values": list[float],
            "user_attr_key": Optional[str],
        },
    )
    SliderWidgetLabel = TypedDict(
        "SliderWidgetLabel",
        {"value": float, "label": str},
    )
    SliderWidgetJSON = TypedDict(
        "SliderWidgetJSON",
        {
            "type": Literal["slider"],
            "description": Optional[str],
            "min": float,
            "max": float,
            "step": Optional[float],
            "labels": Optional[list[SliderWidgetLabel]],
            "user_attr_key": Optional[str],
        },
    )
    TextInputWidgetJSON = TypedDict(
        "TextInputWidgetJSON",
        {
            "type": Literal["text"],
            "description": Optional[str],
            "user_attr_key": Optional[str],
            "optional": bool,
        },
    )
    UserAttrRefJSON = TypedDict("UserAttrRefJSON", {"type": Literal["user_attr"], "key": str})
    FormWidgetJSON = TypedDict(
        "FormWidgetJSON",
        {
            "output_type": Literal["objective", "user_attr"],
            "widgets": Sequence[
                Union[ChoiceWidgetJSON, SliderWidgetJSON, TextInputWidgetJSON, UserAttrRefJSON]
            ],
        },
    )


@dataclass
class ChoiceWidget:
    """A widget representing a choice with associated values.

    Args:
        choices: A list of strings representing the available choices.
        values: A list of float values associated with each choice.
        description: A description of the widget. Defaults to None.
        user_attr_key: The key used by `register_user_attr_form_widgets`.
            Form output is saved as `trial.user_attrs[user_attr_key]`. Defaults to None.

    Example:
        .. code-block:: python

            from optuna_dashboard import ChoiceWidget


            choice_widget = ChoiceWidget(
                choices=["A", "B", "C"], values=[1.0, 2.0, 3.0], description="Choose one"
            )
    """

    choices: list[str]
    values: list[float]
    description: Optional[str] = None
    user_attr_key: Optional[str] = None

    def to_dict(self) -> ChoiceWidgetJSON:
        """Convert the ChoiceWidget object to a dictionary.

        Returns:
            ChoiceWidgetJSON: A dictionary representing the ChoiceWidget object.
        """
        return {
            "type": "choice",
            "description": self.description,
            "choices": self.choices,
            "values": self.values,
            "user_attr_key": self.user_attr_key,
        }

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> ChoiceWidget:
        assert d.get("type") == "choice"
        return cls(
            description=d.get("description"),
            choices=d["choices"],
            values=d["values"],
            user_attr_key=d.get("user_attr_key"),
        )


@dataclass
class SliderWidget:
    """A widget representing a slider for selecting a value within a range.

    Args:
        min: The minimum value of the slider.
        max: The maximum value of the slider.
        step: The step size for the slider. Defaults to None.
        labels: A list of tuples containing value and label for the slider. Defaults to None.
        description: A description for the slider. Defaults to None.
        user_attr_key: The key used by `register_user_attr_form_widgets`.
            Form output is saved as `trial.user_attrs[user_attr_key]`. Defaults to None.

    Example:
        .. code-block:: python

            from optuna_dashboard import SliderWidget


            slide_widget = SliderWidget(min=0, max=10, step=1, description="Example slider")
    """

    min: float
    max: float
    step: Optional[float] = None
    labels: Optional[list[tuple[float, str]]] = None
    description: Optional[str] = None
    user_attr_key: Optional[str] = None

    def to_dict(self) -> SliderWidgetJSON:
        """Convert the SliderWidget instance to a dictionary.

        Returns:
            SliderWidgetJSON: A dictionary representation of the SliderWidget instance.
        """
        labels: Optional[list[SliderWidgetLabel]] = None
        if self.labels is not None:
            labels = [{"value": value, "label": label} for value, label in self.labels]
        return {
            "type": "slider",
            "description": self.description,
            "min": self.min,
            "max": self.max,
            "step": self.step,
            "labels": labels,
            "user_attr_key": self.user_attr_key,
        }

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> SliderWidget:
        assert d.get("type") == "slider"
        labels = d.get("labels")
        if labels is not None:
            labels = [(label["value"], label["label"]) for label in labels]
        return cls(
            description=d.get("description"),
            min=d["min"],
            max=d["max"],
            step=d.get("step"),
            labels=labels,
            user_attr_key=d.get("user_attr_key"),
        )


@dataclass
class TextInputWidget:
    """
    A text input widget class that defines a text input field.

    Args:
        description: A description of the text input field.
        user_attr_key: The key used by `register_user_attr_form_widgets`.
            Form output is saved as `trial.user_attrs[user_attr_key]`. Defaults to None.
        optional: If True, an empty string is acceptable.

    Example:
        .. code-block:: python

            from optuna_dashboard import TextInputWidget


            text_input = TextInputWidget(description="Text Input Example")
    """

    description: Optional[str] = None
    user_attr_key: Optional[str] = None
    optional: bool = False

    def to_dict(self) -> TextInputWidgetJSON:
        """
        Converts the TextInputWidget instance to a dictionary representation.

        Returns:
            TextInputWidgetJSON: The dictionary representation of the TextInputWidget instance.
        """
        return {
            "type": "text",
            "description": self.description,
            "user_attr_key": self.user_attr_key,
            "optional": self.optional,
        }

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> TextInputWidget:
        assert d.get("type") == "text"
        return cls(
            description=d.get("description"),
            user_attr_key=d.get("user_attr_key"),
            optional=d.get("optional", False),
        )


@dataclass
class ObjectiveUserAttrRef:
    """
    A class representing a reference to a value of `trial.user_attrs`.
    When combined with `register_objective_form_widgets`, users can tell values that are
    registered to `trial.user_attrs` during the human-in-the-loop optimization.

    Args:
        key: The key of `trial.user_attrs` being referenced.

    Example:
        .. code-block:: python

            from optuna_dashboard import ObjectiveUserAttrRef


            user_attr_ref = ObjectiveUserAttrRef(key="key")
    """

    key: str

    def to_dict(self) -> UserAttrRefJSON:
        """
        Converts the ObjectiveUserAttrRef instance to a dictionary representation.

        Returns:
            UserAttrRefJSON: The dictionary representation of the ObjectiveUserAttrRef instance.
        """
        return {
            "type": "user_attr",
            "key": self.key,
        }

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> ObjectiveUserAttrRef:
        assert d.get("type") == "user_attr"
        return cls(
            key=d["key"],
        )


ObjectiveFormWidget = Union[ChoiceWidget, SliderWidget, TextInputWidget, ObjectiveUserAttrRef]
# For backward compatibility.
ObjectiveChoiceWidget = ChoiceWidget
ObjectiveSliderWidget = SliderWidget
ObjectiveTextInputWidget = TextInputWidget
FORM_WIDGETS_KEY = "dashboard:form_widgets:v2"


def dict_to_form_widget(d: dict[str, Any]) -> ObjectiveFormWidget:
    """Restore form widget objects from the dictionary.

    Args:
        d: A dictionary object.

    Returns:
        object: an instance of the restored form widget class.
    """
    widget_type = d.get("type", None)
    if widget_type == "choice":
        return ChoiceWidget._from_dict(d)
    elif widget_type == "slider":
        return SliderWidget._from_dict(d)
    elif widget_type == "text":
        return TextInputWidget._from_dict(d)
    elif widget_type == "user_attr":
        return ObjectiveUserAttrRef._from_dict(d)
    raise ValueError("Unexpected widget type")


def register_objective_form_widgets(
    study: optuna.Study, widgets: list[ObjectiveFormWidget]
) -> None:
    """
    Register a list of form widgets to an Optuna study.

    Submitted values to the forms are told as each trial's objective values.

    Args:
        study: The Optuna study object to register the form widgets for.
        widgets: A list of ObjectiveFormWidget objects to be registered in the study.

    Raises:
        ValueError: If the length of study directions is not equal to the length of widgets.
        Warning: If any widget has `user_attr_key` specified, but it will not be used.

    Examples:
        .. code-block:: python

            import optuna
            from optuna_dashboard import ChoiceWidget, SliderWidget
            from optuna_dashboard import register_objective_form_widgets


            study = optuna.create_study()
            register_objective_form_widgets(
                study,
                widgets=[
                    ObjectiveChoiceWidget(
                        choices=["Good 👍", "Bad 👎"],
                        values=[-1, 1],
                        description="Please input your score!",
                    ),
                    ObjectiveSliderWidget(
                        min=1,
                        max=10,
                        step=1,
                        description="Higher is better.",
                    ),
                ],
            )
    """
    if version.parse(optuna_ver) < version.Version("3.2") and isinstance(
        study._storage, optuna.storages._CachedStorage
    ):
        warnings.warn(
            "Human-in-the-loop optimization will not work with _CachedStorage in Optuna prior"
            " to v3.2. See https://optuna-dashboard.readthedocs.io/en/latest/errors.html"
            " for details."
        )

    if len(study.directions) != len(widgets):
        raise ValueError("The length of actions must be the same with the number of objectives.")
    if any(
        not isinstance(w, ObjectiveUserAttrRef) and w.user_attr_key is not None for w in widgets
    ):
        warnings.warn("`user_attr_key` specified, but it will not be used.")
    if any(isinstance(w, TextInputWidget) and w.optional for w in widgets):
        raise ValueError("TextInputWidget.optional must be False.")
    form_widgets: FormWidgetJSON = {
        "output_type": "objective",
        "widgets": [w.to_dict() for w in widgets],
    }
    study._storage.set_study_system_attr(
        study._study_id,
        FORM_WIDGETS_KEY,
        # TODO(c-bata): Remove type: ignore comment
        form_widgets,  # type: ignore
    )


def register_user_attr_form_widgets(
    study: optuna.Study, widgets: list[ObjectiveFormWidget]
) -> None:
    """
    Register a list of form widgets to an Optuna study.

    Submitted values to the forms are registered as each trial's user_attrs.

    Args:
        study: The Optuna study object to register the form widgets for.
        widgets: A list of ObjectiveFormWidget objects to be registered in the study.

    Raises:
        ValueError: If an ObjectiveUserAttrRef is specified or if `user_attr_key` is not specified.
        ValueError: If `user_attr_key` is not unique for each widget.

    Examples:
        .. code-block:: python

            import optuna
            from optuna_dashboard import ChoiceWidget, SliderWidget
            from optuna_dashboard import register_user_attr_form_widgets


            study = optuna.create_study()
            register_user_attr_form_widgets(
                study,
                widgets=[
                    ChoiceWidget(
                        choices=["Good 👍", "Bad 👎"],
                        values=[-1, 1],
                        description="Please input your score!",
                        user_attr_key="hitl/choice",
                    ),
                    SliderWidget(
                        min=1,
                        max=10,
                        step=1,
                        description="Higher is better.",
                        user_attr_key="hitl/slider",
                    ),
                ],
            )
    """
    if version.parse(optuna_ver) < version.Version("3.2") and isinstance(
        study._storage, optuna.storages._CachedStorage
    ):
        warnings.warn(
            "Human-in-the-loop optimization will not work with _CachedStorage in Optuna prior"
            " to v3.2. See https://optuna-dashboard.readthedocs.io/en/latest/errors.html"
            " for details."
        )

    user_attr_keys = set()
    widget_dicts: list[Union[ChoiceWidgetJSON, SliderWidgetJSON, TextInputWidgetJSON]] = []
    for w in widgets:
        if isinstance(w, ObjectiveUserAttrRef):
            raise ValueError("ObjectiveUserAttrRef can't be specified.")
        if w.user_attr_key is None:
            raise ValueError("`user_attr_key` is not specified.")
        user_attr_keys.add(w.user_attr_key)
        widget_dicts.append(w.to_dict())

    if len(widget_dicts) != len(user_attr_keys):
        raise ValueError("`user_attr_key` must be unique for each widget.")

    form_widgets: FormWidgetJSON = {
        "output_type": "user_attr",
        "widgets": widget_dicts,
    }
    study._storage.set_study_system_attr(
        study._study_id,
        FORM_WIDGETS_KEY,
        # TODO(c-bata): Remove type: ignore comment
        form_widgets,  # type: ignore
    )


def get_form_widgets_json(study_system_attr: dict[str, Any]) -> Optional[FormWidgetJSON]:
    if FORM_WIDGETS_KEY in study_system_attr:
        return study_system_attr[FORM_WIDGETS_KEY]

    # For optuna-dashboard v0.9.0 and v0.9.0b6 users
    if "dashboard:objective_form_widgets:v1" in study_system_attr:
        return {
            "output_type": "objective",
            "widgets": study_system_attr["dashboard:objective_form_widgets:v1"],
        }

    # For optuna-dashboard v0.9.0b5 users
    if "dashboard:objective_form_widgets" in study_system_attr:
        return {
            "output_type": "objective",
            "widgets": json.loads(study_system_attr["dashboard:objective_form_widgets"]),
        }
    return None



from __future__ import annotations

import io
import json
import mimetypes
import os.path
from typing import TYPE_CHECKING
import uuid
import warnings

from bottle import BaseRequest
from bottle import Bottle
from bottle import HTTPResponse
from bottle import request
from bottle import response
import optuna
from optuna.trial import FrozenTrial

from .._bottle_util import json_api_view
from .._bottle_util import parse_data_uri


if TYPE_CHECKING:
    from typing import Any
    from typing import Optional
    from typing import TypedDict

    from optuna.artifacts._protocol import ArtifactStore
    from optuna.storages import BaseStorage

    from .protocol import ArtifactBackend

    ArtifactMeta = TypedDict(
        "ArtifactMeta",
        {
            "artifact_id": str,
            "filename": str,
            "mimetype": str,
            "encoding": Optional[str],
        },
    )

ARTIFACTS_ATTR_PREFIX = "artifacts:"
DASHBOARD_ARTIFACTS_ATTR_PREFIX = "dashboard:artifacts:"
DEFAULT_MIME_TYPE = "application/octet-stream"
BaseRequest.MEMFILE_MAX = int(
    os.environ.get("OPTUNA_DASHBOARD_MEMFILE_MAX", 1024 * 1024 * 128)
)  # 128MB


def get_artifact_path(
    study_or_trial: optuna.Trial | optuna.Study,
    artifact_id: str,
) -> str:
    """Get the URL path for a given artifact ID.

    Args:
        study_or_trial:
            A Trial object, or a Study object.

        artifact_id:
            An artifact ID.

    Returns:
        A URL path to the artifact.
    """
    if isinstance(study_or_trial, optuna.Study):
        study_id = study_or_trial._study_id
        return f"/artifacts/{study_id}/{artifact_id}"

    study_id = study_or_trial.study._study_id
    trial_id = study_or_trial._trial_id
    return f"/artifacts/{study_id}/{trial_id}/{artifact_id}"


def register_artifact_route(
    app: Bottle, storage: BaseStorage, artifact_store: ArtifactStore | None
) -> None:
    @app.get("/artifacts/<study_id:int>/<artifact_id:re:[0-9a-fA-F-]+>")
    def proxy_study_artifact(study_id: int, artifact_id: str) -> HTTPResponse | bytes:
        if artifact_store is None:
            response.status = 400  # Bad Request
            return b"Cannot access to the artifacts."
        artifact_dict = get_study_artifact_meta(storage, study_id, artifact_id)
        if artifact_dict is None:
            response.status = 404
            return b"Not Found"
        headers = {"Content-Type": artifact_dict["mimetype"]}
        encoding = artifact_dict.get("encoding")
        if encoding:
            headers["Content-Encodings"] = encoding

        # CORS Headers
        headers["Access-Control-Allow-Origin"] = "*"

        fp = artifact_store.open_reader(artifact_id)
        return HTTPResponse(fp, headers=headers)

    @app.get("/artifacts/<study_id:int>/<trial_id:int>/<artifact_id:re:[0-9a-fA-F-]+>")
    def proxy_trial_artifact(
        study_id: int,
        trial_id: int,
        artifact_id: str,
    ) -> HTTPResponse | bytes:
        if artifact_store is None:
            response.status = 400  # Bad Request
            return b"Cannot access to the artifacts."
        artifact_dict = get_trial_artifact_meta(storage, study_id, trial_id, artifact_id)
        if artifact_dict is None:
            response.status = 404
            return b"Not Found"
        headers = {"Content-Type": artifact_dict["mimetype"]}
        encoding = artifact_dict.get("encoding")
        if encoding:
            headers["Content-Encodings"] = encoding

        # CORS Headers
        headers["Access-Control-Allow-Origin"] = "*"

        fp = artifact_store.open_reader(artifact_id)
        return HTTPResponse(fp, headers=headers)

    @app.post("/api/artifacts/<study_id:int>/<trial_id:int>")
    @json_api_view
    def upload_trial_artifact_api(study_id: int, trial_id: int) -> dict[str, Any]:
        trial = storage.get_trial(trial_id)
        if trial is None:
            response.status = 400
            return {"reason": "Invalid study_id or trial_id"}
        elif trial.state.is_finished():
            response.status = 400
            return {"reason": "The trial is already finished."}

        # TODO(c-bata): Use optuna.artifacts.upload_artifact()
        if artifact_store is None:
            response.status = 400  # Bad Request
            return {"reason": "Cannot access to the artifacts."}
        file = request.json.get("file")
        if file is None:
            response.status = 400
            return {"reason": "Please specify the 'file' key."}

        _, data = parse_data_uri(file)
        filename = request.json.get("filename", "")
        artifact_id = str(uuid.uuid4())
        artifact_store.write(artifact_id, io.BytesIO(data))

        mimetype, encoding = mimetypes.guess_type(filename)
        artifact = {
            "artifact_id": artifact_id,
            "filename": filename,
            "mimetype": mimetype or DEFAULT_MIME_TYPE,
            "encoding": encoding,
        }
        attr_key = ARTIFACTS_ATTR_PREFIX + artifact_id
        storage.set_trial_system_attr(trial_id, attr_key, json.dumps(artifact))
        response.status = 201

        study_system_attrs = storage.get_study_system_attrs(study_id)
        trial_system_attrs = storage.get_trial_system_attrs(trial_id)
        artifacts = list_trial_artifacts(study_system_attrs, trial_system_attrs, trial)
        return {
            "artifact_id": artifact_id,
            "artifacts": artifacts,
        }

    @app.post("/api/artifacts/<study_id:int>")
    @json_api_view
    def upload_study_artifact_api(study_id: int) -> dict[str, Any]:
        if artifact_store is None:
            response.status = 400  # Bad Request
            return {"reason": "Cannot access to the artifacts."}
        file = request.json.get("file")
        if file is None:
            response.status = 400
            return {"reason": "Please specify the 'file' key."}

        _, data = parse_data_uri(file)
        filename = request.json.get("filename", "")
        artifact_id = str(uuid.uuid4())
        artifact_store.write(artifact_id, io.BytesIO(data))

        mimetype, encoding = mimetypes.guess_type(filename)
        artifact = {
            "artifact_id": artifact_id,
            "filename": filename,
            "mimetype": mimetype or DEFAULT_MIME_TYPE,
            "encoding": encoding,
        }
        attr_key = ARTIFACTS_ATTR_PREFIX + artifact_id
        storage.set_study_system_attr(study_id, attr_key, json.dumps(artifact))

        response.status = 201

        return {
            "artifact_id": artifact_id,
            "artifacts": list_study_artifacts(storage.get_study_system_attrs(study_id)),
        }

    @app.delete("/api/artifacts/<study_id:int>/<trial_id:int>/<artifact_id:re:[0-9a-fA-F-]+>")
    @json_api_view
    def delete_trial_artifact(study_id: int, trial_id: int, artifact_id: str) -> dict[str, Any]:
        from optuna.artifacts.exceptions import ArtifactNotFound

        if artifact_store is None:
            response.status = 400  # Bad Request
            return {"reason": "Cannot access to the artifacts."}
        try:
            artifact_store.remove(artifact_id)
        except ArtifactNotFound as e:
            response.status = 404
            return {"reason": str(e)}

        # The artifact's metadata is stored in one of the following two locations:
        storage.set_study_system_attr(
            study_id, _dashboard_artifact_prefix(trial_id) + artifact_id, json.dumps(None)
        )
        storage.set_trial_system_attr(
            trial_id, ARTIFACTS_ATTR_PREFIX + artifact_id, json.dumps(None)
        )

        response.status = 204
        return {}

    @app.delete("/api/artifacts/<study_id:int>/<artifact_id:re:[0-9a-fA-F-]+>")
    @json_api_view
    def delete_study_artifact(study_id: int, artifact_id: str) -> dict[str, Any]:
        from optuna.artifacts.exceptions import ArtifactNotFound

        if artifact_store is None:
            response.status = 400  # Bad Request
            return {"reason": "Cannot access to the artifacts."}
        try:
            artifact_store.remove(artifact_id)
        except ArtifactNotFound as e:
            response.status = 404
            return {"reason": str(e)}

        storage.set_study_system_attr(
            study_id, ARTIFACTS_ATTR_PREFIX + artifact_id, json.dumps(None)
        )

        response.status = 204
        return {}


def upload_artifact(
    backend: ArtifactBackend,
    trial: optuna.Trial,
    file_path: str,
    *,
    mimetype: str | None = None,
    encoding: str | None = None,
) -> str:
    """Upload an artifact (files), which is associated with the trial.

    .. warning::

       This function is deprecated. Please use `optuna.artifacts.upload_artifact
       <https://optuna.readthedocs.io/en/latest/reference/generated/optuna.artifacts.
       upload_artifact.html>`_ instead.

    Example:
       .. code-block:: python

          import optuna
          from optuna_dashboard.artifact import upload_artifact
          from optuna_dashboard.artifact.file_system import FileSystemBackend

          artifact_backend = FileSystemBackend("./tmp/")

          def objective(trial: optuna.Trial) -> float:
              ... = trial.suggest_float("x", -10, 10)
              file_path = generate_example_png(...)
              upload_artifact(artifact_backend, trial, file_path)
              return ...
    """
    warnings.warn(
        "upload_artifact() is deprecated. Please use optuna.artifacts.upload_artifact() instead.\n"
        "See https://optuna-dashboard.readthedocs.io/en/latest/errors.html for details",
        DeprecationWarning,
    )

    filename = os.path.basename(file_path)
    storage = trial.storage
    trial_id = trial._trial_id
    artifact_id = str(uuid.uuid4())
    guess_mimetype, guess_encoding = mimetypes.guess_type(filename)
    artifact: ArtifactMeta = {
        "artifact_id": artifact_id,
        "mimetype": mimetype or guess_mimetype or DEFAULT_MIME_TYPE,
        "encoding": encoding or guess_encoding,
        "filename": filename,
    }
    attr_key = ARTIFACTS_ATTR_PREFIX + artifact_id
    storage.set_trial_system_attr(trial_id, attr_key, json.dumps(artifact))

    with open(file_path, "rb") as f:
        backend.write(artifact_id, f)
    return artifact_id


def _dashboard_artifact_prefix(trial_id: int) -> str:
    return DASHBOARD_ARTIFACTS_ATTR_PREFIX + f"{trial_id}:"


def get_study_artifact_meta(
    storage: BaseStorage, study_id: int, artifact_id: str
) -> ArtifactMeta | None:
    study_system_attrs = storage.get_study_system_attrs(study_id)
    attr_key = ARTIFACTS_ATTR_PREFIX + artifact_id
    artifact_meta = study_system_attrs.get(attr_key)
    if artifact_meta is not None:
        return json.loads(artifact_meta)
    return None


def get_trial_artifact_meta(
    storage: BaseStorage, study_id: int, trial_id: int, artifact_id: str
) -> ArtifactMeta | None:
    # Search study_system_attrs due to backward compatibility.
    study_system_attrs = storage.get_study_system_attrs(study_id)
    attr_key = _dashboard_artifact_prefix(trial_id=trial_id) + artifact_id
    artifact_meta = study_system_attrs.get(attr_key)
    if artifact_meta is not None:
        return json.loads(artifact_meta)

    # Search trial_system_attrs. Note that artifacts uploaded via optuna.artifacts.upload_artifact
    # have a different trial_system_attrs key prefix.
    # See https://github.com/optuna/optuna/blob/f827582a8/optuna/artifacts/_upload.py#L71
    trial_system_attrs = storage.get_trial_system_attrs(trial_id)
    value = trial_system_attrs.get(ARTIFACTS_ATTR_PREFIX + artifact_id)
    if value is not None:
        return json.loads(value)

    return None


def delete_all_artifacts(backend: ArtifactStore, storage: BaseStorage, study_id: int) -> None:
    artifact_metas = []
    study_system_attrs = storage.get_study_system_attrs(study_id)
    artifact_metas.extend(list_study_artifacts(study_system_attrs))
    for trial in storage.get_all_trials(study_id):
        trial_system_attrs = getattr(trial, "_system_attrs")
        if trial_system_attrs is None:
            # This is unreachable line until Optuna v5.0.0 release.
            trial_system_attrs = storage.get_trial_system_attrs(trial._trial_id)
        trial_artifacts = list_trial_artifacts(study_system_attrs, trial_system_attrs, trial)
        artifact_metas.extend(trial_artifacts)

    for meta in artifact_metas:
        backend.remove(meta["artifact_id"])


def list_study_artifacts(study_system_attrs: dict[str, Any]) -> list[ArtifactMeta]:
    artifact_metas = [
        json.loads(value)
        for key, value in study_system_attrs.items()
        if key.startswith(ARTIFACTS_ATTR_PREFIX)
    ]
    return [a for a in artifact_metas if a is not None]


def list_trial_artifacts(
    study_system_attrs: dict[str, Any], trial_system_attrs: dict[str, Any], trial: FrozenTrial
) -> list[ArtifactMeta]:
    # Collect ArtifactMeta from study_system_attrs due to backward compatibility.
    dashboard_artifact_metas = [
        json.loads(value)
        for key, value in study_system_attrs.items()
        if key.startswith(_dashboard_artifact_prefix(trial._trial_id))
    ]

    # Collect ArtifactMeta from trial_system_attrs. Note that artifacts uploaded via
    # optuna.artifacts.upload_artifacts have a different trial_system_attrs key prefix.
    # See https://github.com/optuna/optuna/blob/f827582a8/optuna/artifacts/_upload.py#L16
    optuna_artifact_metas = [
        json.loads(value)
        for key, value in trial_system_attrs.items()
        if key.startswith(ARTIFACTS_ATTR_PREFIX)
    ]
    artifact_metas = dashboard_artifact_metas + optuna_artifact_metas
    return [a for a in artifact_metas if a is not None]



from __future__ import annotations

from datetime import datetime
import json
import numbers
from typing import Any
from typing import TYPE_CHECKING
from typing import Union

import numpy as np
from optuna.distributions import BaseDistribution
from optuna.distributions import CategoricalDistribution
from optuna.distributions import FloatDistribution
from optuna.distributions import IntDistribution
from optuna.study._frozen import FrozenStudy
from optuna.trial import FrozenTrial

from . import _note as note
from ._form_widget import get_form_widgets_json
from ._named_objectives import get_objective_names
from ._preference_setting import _SYSTEM_ATTR_FEEDBACK_COMPONENT
from ._preferential_history import _SYSTEM_ATTR_PREFIX_HISTORY
from .artifact._backend import list_study_artifacts
from .artifact._backend import list_trial_artifacts
from .preferential._study import _SYSTEM_ATTR_PREFERENTIAL_STUDY
from .preferential._system_attrs import get_preferences
from .preferential._system_attrs import is_preference_removed


if TYPE_CHECKING:
    from typing import Literal
    from typing import TypedDict

    from ._preferential_history import History
    from ._preferential_history import SerializedHistory

    Attribute = TypedDict(
        "Attribute",
        {
            "key": str,
            "value": str,
        },
    )
    AttributeSpec = TypedDict(
        "AttributeSpec",
        {
            "key": str,
            "sortable": bool,
        },
    )
    IntermediateValue = TypedDict(
        "IntermediateValue",
        {
            "step": int,
            "value": Union[float, Literal["inf", "-inf", "nan"]],
        },
    )

    FloatDistributionJSON = TypedDict(
        "FloatDistributionJSON",
        {
            "type": Literal["FloatDistribution"],
            "low": float,
            "high": float,
            "step": Union[float, None],
            "log": bool,
        },
    )
    IntDistributionJSON = TypedDict(
        "IntDistributionJSON",
        {
            "type": Literal["IntDistribution"],
            "low": int,
            "high": int,
            "step": int,
            "log": bool,
        },
    )
    CategoricalDistributionChoiceJSON = TypedDict(
        "CategoricalDistributionChoiceJSON",
        {
            "pytype": str,
            "value": str,
        },
    )
    CategoricalDistributionJSON = TypedDict(
        "CategoricalDistributionJSON",
        {
            "type": Literal["CategoricalDistribution"],
            "choices": list[CategoricalDistributionChoiceJSON],
        },
    )
    DistributionJSON = Union[
        FloatDistributionJSON, IntDistributionJSON, CategoricalDistributionJSON
    ]


MAX_ATTR_LENGTH = 1024
CONSTRAINTS_KEY = "constraints"


def serialize_attrs(attrs: dict[str, Any]) -> list[Attribute]:
    serialized: list[Attribute] = []
    for k, v in attrs.items():
        value: str
        if isinstance(v, bytes):
            value = "<binary object>"
        elif isinstance(v, str):
            value = v
        elif isinstance(v, numbers.Real):
            value = str(v)
        else:
            value = json.dumps(v)
            value = value[:MAX_ATTR_LENGTH] if len(value) > MAX_ATTR_LENGTH else value
        serialized.append({"key": k, "value": value})
    return serialized


def serialize_frozen_study(study: FrozenStudy) -> dict[str, Any]:
    serialized = {
        "study_id": study._study_id,
        "study_name": study.study_name,
        "directions": [d.name.lower() for d in study.directions],
        "user_attrs": serialize_attrs(study.user_attrs),
        "is_preferential": study.system_attrs.get(_SYSTEM_ATTR_PREFERENTIAL_STUDY, False),
    }

    return serialized


def serialize_study_detail(
    study: FrozenStudy,
    best_trials: list[FrozenTrial],
    trials: list[FrozenTrial],
    intersection: list[tuple[str, BaseDistribution]],
    union: list[tuple[str, BaseDistribution]],
    union_user_attrs: list[tuple[str, bool]],
    has_intermediate_values: bool,
    plotly_graph_objects: dict[str, str],
    skipped_trial_numbers: list[int],
) -> dict[str, Any]:
    serialized: dict[str, Any] = {
        "name": study.study_name,
        "directions": [d.name.lower() for d in study.directions],
        "user_attrs": serialize_attrs(study.user_attrs),
    }
    system_attrs = study.system_attrs
    serialized["artifacts"] = list_study_artifacts(system_attrs)

    serialized["trials"] = [
        serialize_frozen_trial(study._study_id, trial, system_attrs) for trial in trials
    ]
    serialized["best_trials"] = [
        serialize_frozen_trial(study._study_id, trial, system_attrs) for trial in best_trials
    ]
    serialized["intersection_search_space"] = serialize_search_space(intersection)
    serialized["union_search_space"] = serialize_search_space(union)
    serialized["union_user_attrs"] = [{"key": a[0], "sortable": a[1]} for a in union_user_attrs]
    serialized["has_intermediate_values"] = has_intermediate_values
    serialized["note"] = note.get_note_from_system_attrs(system_attrs, None)
    serialized["is_preferential"] = system_attrs.get(_SYSTEM_ATTR_PREFERENTIAL_STUDY, False)
    objective_names = get_objective_names(system_attrs)
    if objective_names:
        serialized["objective_names"] = objective_names
    form_widgets = get_form_widgets_json(system_attrs)
    if form_widgets:
        serialized["form_widgets"] = form_widgets
    serialized["feedback_component_type"] = system_attrs.get(
        _SYSTEM_ATTR_FEEDBACK_COMPONENT,
        {
            "output_type": "note",
        },
    )
    if serialized["is_preferential"]:
        serialized["preference_history"] = serialize_preference_history(system_attrs)
        serialized["preferences"] = get_preferences(system_attrs)
        serialized["skipped_trial_numbers"] = skipped_trial_numbers
    serialized["plotly_graph_objects"] = [
        {"id": id_, "graph_object": graph_object}
        for id_, graph_object in plotly_graph_objects.items()
    ]
    return serialized


def serialize_preference_history(
    system_attrs: dict[str, Any],
) -> list[SerializedHistory]:
    histories: list[SerializedHistory] = []
    for k, v in system_attrs.items():
        if not k.startswith(_SYSTEM_ATTR_PREFIX_HISTORY):
            continue
        choice: dict[str, Any] = json.loads(v)
        if choice["mode"] == "ChooseWorst":
            history: History = {
                "mode": "ChooseWorst",
                "id": choice["id"],
                "timestamp": choice["timestamp"],
                "candidates": choice["candidates"],
                "clicked": choice["clicked"],
                "preferences": choice["preferences"],
            }
            histories.append(
                {
                    "history": history,
                    "is_removed": is_preference_removed(system_attrs, choice["id"]),
                }
            )

    histories.sort(key=lambda c: datetime.fromisoformat(c["history"]["timestamp"]))
    return histories


def serialize_frozen_trial(
    study_id: int, trial: FrozenTrial, study_system_attrs: dict[str, Any]
) -> dict[str, Any]:
    params = []
    for param_name, param_external_value in trial.params.items():
        distribution = trial.distributions.get(param_name)
        if distribution is None:
            continue
        params.append(
            {
                "name": param_name,
                "param_internal_value": distribution.to_internal_repr(param_external_value),
                "param_external_value": str(param_external_value),
                "param_external_pytyp": str(type(param_external_value)),
                "distribution": serialize_distribution(distribution),
            }
        )
    trial_system_attrs: dict[str, Any] = getattr(trial, "_system_attrs", {})
    fixed_params = trial_system_attrs.get("fixed_params", {})
    serialized = {
        "trial_id": trial._trial_id,
        "study_id": study_id,
        "number": trial.number,
        "state": trial.state.name.capitalize(),
        "params": params,
        "fixed_params": [
            {"name": param_name, "param_external_value": str(fixed_params.get(param_name, None))}
            for param_name in fixed_params
        ],
        "user_attrs": serialize_attrs(trial.user_attrs),
        "note": note.get_note_from_system_attrs(study_system_attrs, trial._trial_id),
        "artifacts": list_trial_artifacts(study_system_attrs, trial_system_attrs, trial),
        "constraints": trial_system_attrs.get(CONSTRAINTS_KEY, []),
    }

    serialized_intermediate_values: list[IntermediateValue] = []
    for step, value in trial.intermediate_values.items():
        serialized_value: Union[float, Literal["nan", "inf", "-inf"]]
        if np.isnan(value):
            serialized_value = "nan"
        elif np.isposinf(value):
            serialized_value = "inf"
        elif np.isneginf(value):
            serialized_value = "-inf"
        else:
            assert np.isfinite(value)
            serialized_value = value
        serialized_intermediate_values.append({"step": step, "value": serialized_value})
    serialized["intermediate_values"] = sorted(
        serialized_intermediate_values, key=lambda v: v["step"]
    )

    if trial.values is not None:
        serialized_values: list[Union[float, Literal["inf", "-inf"]]] = []
        for v in trial.values:
            assert not np.isnan(v), "Should not detect nan value"
            if np.isposinf(v):
                serialized_values.append("inf")
            elif np.isneginf(v):
                serialized_values.append("-inf")
            else:
                serialized_values.append(v)
        serialized["values"] = serialized_values

    if trial.datetime_start is not None:
        serialized["datetime_start"] = trial.datetime_start.isoformat()

    if trial.datetime_complete is not None:
        serialized["datetime_complete"] = trial.datetime_complete.isoformat()

    return serialized


def serialize_distribution(distribution: BaseDistribution) -> DistributionJSON:
    if isinstance(distribution, FloatDistribution):
        float_distribution: FloatDistributionJSON = {
            "type": "FloatDistribution",
            "low": getattr(distribution, "low"),
            "high": getattr(distribution, "high"),
            "step": getattr(distribution, "step"),
            "log": getattr(distribution, "log"),
        }
        return float_distribution
    if isinstance(distribution, IntDistribution):
        int_distribution: IntDistributionJSON = {
            "type": "IntDistribution",
            "low": getattr(distribution, "low"),
            "high": getattr(distribution, "high"),
            "step": getattr(distribution, "step"),
            "log": getattr(distribution, "log"),
        }
        return int_distribution
    if isinstance(distribution, CategoricalDistribution):
        categorical: CategoricalDistributionJSON = {
            "type": "CategoricalDistribution",
            "choices": [
                {"pytype": str(type(choice)), "value": str(choice)}
                for choice in distribution.choices
            ],
        }
        return categorical
    raise ValueError(f"Unexpected distribution {str(distribution)}")


def serialize_search_space(
    search_space: list[tuple[str, BaseDistribution]],
) -> list[dict[str, Any]]:
    serialized = []
    for param_name, distribution in search_space:
        serialized.append(
            {
                "name": param_name,
                "distribution": serialize_distribution(distribution),
            }
        )
    return serialized



from __future__ import annotations


class ArtifactNotFound(Exception):
    """Exception raised when an artifact is not found.

    It is typically raised while calling
    :meth:`~optuna_dashboard.artifact.protocol.ArtifactBackend.open` or
    :meth:`~optuna_dashboard.artifact.protocol.ArtifactBackend.remove` methods.
    """

    ...



from __future__ import annotations

import csv
from dataclasses import dataclass
import functools
import importlib
import io
from itertools import chain
import logging
import mimetypes
import os
import re
import typing
import warnings

from bottle import Bottle
from bottle import redirect
from bottle import request
from bottle import response
from bottle import run
from bottle import static_file
import optuna
from optuna.exceptions import DuplicatedStudyError
from optuna.storages import BaseStorage
from optuna.study import StudyDirection
from optuna.trial import TrialState

from . import _note as note
from ._bottle_util import BottleViewReturn
from ._bottle_util import json_api_view
from ._custom_plot_data import get_plotly_graph_objects
from ._importance import get_param_importance_from_trials_cache
from ._inmemory_cache import get_cached_extra_study_property
from ._inmemory_cache import InMemoryCache
from ._pareto_front import get_pareto_front_trials
from ._preference_setting import _register_preference_feedback_component
from ._preferential_history import NewHistory
from ._preferential_history import PreferenceHistoryNotFound
from ._preferential_history import remove_history
from ._preferential_history import report_history
from ._preferential_history import restore_history
from ._rdb_migration import register_rdb_migration_route
from ._serializer import serialize_frozen_study
from ._serializer import serialize_study_detail
from ._storage import create_new_study
from ._storage import get_studies
from ._storage import get_study
from ._storage import get_trials
from ._storage_url import get_storage
from .artifact._backend import delete_all_artifacts
from .artifact._backend import register_artifact_route
from .artifact._backend_to_store import to_artifact_store
from .preferential._study import _SYSTEM_ATTR_PREFERENTIAL_STUDY
from .preferential._study import get_best_trials as get_best_preferential_trials
from .preferential._system_attrs import get_skipped_trial_ids
from .preferential._system_attrs import report_skip


if typing.TYPE_CHECKING:
    from typing import Any
    from typing import Literal
    from typing import Optional
    from typing import Union

    from _typeshed.wsgi import WSGIApplication
    from optuna.artifacts._protocol import ArtifactStore
    from optuna_dashboard.artifact.protocol import ArtifactBackend


logger = logging.getLogger(__name__)

# Static files
BASE_DIR = os.path.dirname(os.path.abspath(__file__))
STATIC_DIR = os.path.join(BASE_DIR, "public")
IMG_DIR = os.path.join(BASE_DIR, "img")
cached_path_exists = functools.lru_cache(maxsize=10)(os.path.exists)


@dataclass
class JupyterLabExtensionContext:
    base_url: str


def create_app(
    storage: BaseStorage,
    artifact_store: Optional[ArtifactStore] = None,
    debug: bool = False,
    jupyterlab_extension_context: JupyterLabExtensionContext | None = None,
) -> Bottle:
    app = Bottle()
    app._inmemory_cache = InMemoryCache()

    @app.hook("before_request")
    def remove_trailing_slashes_hook() -> None:
        request.environ["PATH_INFO"] = request.environ["PATH_INFO"].rstrip("/")

    @app.get("/")
    def index() -> BottleViewReturn:
        return redirect("/dashboard", 302)  # Status Found

    # Accept any following paths for client-side routing
    @app.get("/dashboard<:re:(/.*)?>")
    def dashboard() -> BottleViewReturn:
        return static_file("index.html", BASE_DIR, mimetype="text/html")

    @app.get("/api/meta")
    @json_api_view
    def api_meta() -> dict[str, Any]:
        meta: dict[str, Any] = {
            "artifact_is_available": artifact_store is not None,
            "plotlypy_is_available": importlib.util.find_spec("plotly") is not None,
        }
        if jupyterlab_extension_context is not None:
            meta["jupyterlab_extension_context"] = {
                "base_url": jupyterlab_extension_context.base_url
            }
        return meta

    @app.get("/api/studies")
    @json_api_view
    def list_studies() -> dict[str, Any]:
        studies = get_studies(storage)
        serialized = [serialize_frozen_study(s) for s in studies]
        # TODO(umezawa): Rename `study_summaries` to `studies`.
        return {
            "study_summaries": serialized,
        }

    @app.post("/api/studies")
    @json_api_view
    def create_study() -> dict[str, Any]:
        study_name = request.json.get("study_name", None)
        request_directions = [d.lower() for d in request.json.get("directions", [])]
        if (
            study_name is None
            or len(request_directions) == 0
            or not all([d in ("minimize", "maximize") for d in request_directions])
        ):
            response.status = 400  # Bad request
            return {"reason": "You need to set study_name and direction"}

        directions = [
            StudyDirection.MAXIMIZE if d == "maximize" else StudyDirection.MINIMIZE
            for d in request_directions
        ]
        try:
            study_id = create_new_study(storage, study_name, directions)
        except DuplicatedStudyError:
            response.status = 400  # Bad request
            return {"reason": f"'{study_name}' already exists"}

        study = get_study(storage, study_id)
        if study is None:
            response.status = 500  # Internal server error
            return {"reason": "Failed to create study"}
        response.status = 201  # Created
        return {"study_summary": serialize_frozen_study(study)}

    @app.post("/api/studies/<study_id:int>/rename")
    @json_api_view
    def rename_study(study_id: int) -> dict[str, Any]:
        dst_study_name = request.json.get("study_name", None)
        if dst_study_name is None:
            response.status = 400  # Bad request
            return {"reason": "You need to set study_name and direction"}

        src_study_name = storage.get_study_name_from_id(study_id)
        try:
            src_study = optuna.load_study(storage=storage, study_name=src_study_name)
        except KeyError:
            response.status = 404  # Not found
            return {"reason": f"study_id={study_id} is not found"}

        try:
            dst_study = optuna.create_study(
                storage=storage, study_name=dst_study_name, directions=src_study.directions
            )
            dst_study.add_trials(src_study.get_trials(deepcopy=False))
            for key, value in src_study.user_attrs.items():
                dst_study.set_user_attr(key, value)
            note.copy_notes(storage, src_study, dst_study)
        except DuplicatedStudyError:
            response.status = 400  # Bad request
            return {"reason": f"study_name={dst_study_name} is duplicaated"}
        except Exception as e:
            logger.exception("Unexpected error:")
            response.status = 500
            storage.delete_study(dst_study._study_id)
            return {"reason": str(e)}
        new_study = get_study(storage, dst_study._study_id)
        if new_study is None:
            response.status = 500
            return {"reason": "Failed to load the new study"}

        storage.delete_study(src_study._study_id)
        response.status = 201
        return serialize_frozen_study(new_study)

    @app.delete("/api/studies/<study_id:int>")
    @json_api_view
    def delete_study(study_id: int) -> dict[str, Any]:
        data = request.json or {}
        remove_associated_artifacts = data.get("remove_associated_artifacts", True)

        if artifact_store is not None and remove_associated_artifacts:
            delete_all_artifacts(artifact_store, storage, study_id)

        try:
            storage.delete_study(study_id)
        except KeyError:
            response.status = 404  # Not found
            return {"reason": f"study_id={study_id} is not found"}
        response.status = 204  # No content
        return {}

    @app.get("/api/studies/<study_id:int>")
    @json_api_view
    def get_study_detail(study_id: int) -> dict[str, Any]:
        try:
            after = int(request.params["after"])
            assert after >= 0
        except AssertionError:
            response.status = 400  # Bad parameter
            return {"reason": "`after` should be larger or equal 0."}
        except KeyError:
            after = 0
        study = get_study(storage, study_id)
        if study is None:
            response.status = 404  # Not found
            return {"reason": f"study_id={study_id} is not found"}
        trials = get_trials(app._inmemory_cache, storage, study_id)

        system_attrs = getattr(study, "system_attrs", {})
        is_preferential = system_attrs.get(_SYSTEM_ATTR_PREFERENTIAL_STUDY, False)
        # TODO(c-bata): Cache best_trials
        if is_preferential:
            best_trials = get_best_preferential_trials(study_id, storage)
        elif len(study.directions) == 1:
            if len([t for t in trials if t.state == TrialState.COMPLETE]) == 0:
                best_trials = []
            else:
                best_trials = [storage.get_best_trial(study_id)]
        else:
            best_trials = get_pareto_front_trials(trials=trials, directions=study.directions)
        (
            # TODO: intersection_search_space and union_search_space look more clear since now we
            # have union_user_attrs.
            intersection,
            union,
            union_user_attrs,
            has_intermediate_values,
        ) = get_cached_extra_study_property(app._inmemory_cache, study_id, trials)

        plotly_graph_objects = get_plotly_graph_objects(system_attrs)
        skipped_trial_ids = get_skipped_trial_ids(system_attrs)
        skipped_trial_numbers = [t.number for t in trials if t._trial_id in skipped_trial_ids]
        return serialize_study_detail(
            study,
            best_trials,
            trials[after:],
            intersection,
            union,
            union_user_attrs,
            has_intermediate_values,
            plotly_graph_objects,
            skipped_trial_numbers,
        )

    @app.get("/api/studies/<study_id:int>/param_importances")
    @json_api_view
    def get_param_importances(study_id: int) -> dict[str, Any]:
        try:
            n_directions = len(storage.get_study_directions(study_id))
        except KeyError:
            response.status = 404  # Study is not found
            return {"reason": f"study_id={study_id} is not found"}

        trials = get_trials(app._inmemory_cache, storage, study_id)

        evaluator = request.query.get("evaluator", "ped_anova")  # Default evaluator is "ped_anova"
        if (
            evaluator != "fanova"
            and evaluator != "ped_anova"
            and evaluator != "mean_decrease_impurity"
        ):
            response.status = 400
            return {
                "reason": (
                    "evaluator must be either 'fanova', 'ped_anova' or 'mean_decrease_impurity'."
                )
            }

        try:
            importances = [
                get_param_importance_from_trials_cache(
                    app._inmemory_cache,
                    storage,
                    study_id,
                    objective_id,
                    evaluator,
                    trials,
                )
                for objective_id in range(n_directions)
            ]
            return {"param_importances": importances}
        except ValueError as e:
            response.status = 400  # Bad request
            return {"reason": str(e)}

    @app.get("/api/studies/<study_id:int>/plot/<plot_type>")
    @json_api_view
    def get_plot(study_id: int, plot_type: str) -> dict[str, Any]:
        study = optuna.load_study(
            study_name=storage.get_study_name_from_id(study_id), storage=storage
        )
        if plot_type == "contour":
            fig = optuna.visualization.plot_contour(study)
        elif plot_type == "slice":
            fig = optuna.visualization.plot_slice(study)
            # Note: Optuna's implementation forces a minimum width.
            # We override it to prevent the figure from going beyond the screen width.
            # https://github.com/optuna/optuna/blob/2abd0ae81eaf3683ce1dd580429904c8a705300d/optuna/visualization/_slice.py#L237-L239
            fig.update_layout(width=None)
        elif plot_type == "parallel_coordinate":
            fig = optuna.visualization.plot_parallel_coordinate(study)
        elif plot_type == "rank":
            fig = optuna.visualization.plot_rank(study)
        elif plot_type == "edf":
            fig = optuna.visualization.plot_edf(study)
        elif plot_type == "timeline":
            fig = optuna.visualization.plot_timeline(study)
        elif plot_type == "param_importances":
            fig = optuna.visualization.plot_param_importances(study)
        elif plot_type == "pareto_front":
            fig = optuna.visualization.plot_pareto_front(study)
        else:
            response.status = 404  # Not found
            return {"reason": f"plot_type={plot_type} is not supported."}
        return fig.to_json()

    @app.get("/api/compare-studies/plot/<plot_type>")
    @json_api_view
    def get_compare_studies_plot(plot_type: str) -> dict[str, Any]:
        study_ids = map(int, request.query.getall("study_ids[]"))
        studies = [
            optuna.load_study(study_name=storage.get_study_name_from_id(study_id), storage=storage)
            for study_id in study_ids
        ]
        if plot_type == "edf":
            fig = optuna.visualization.plot_edf(studies)
        else:
            response.status = 404  # Not found
            return {"reason": f"plot_type={plot_type} is not supported."}
        return fig.to_json()

    @app.put("/api/studies/<study_id:int>/note")
    @json_api_view
    def save_study_note(study_id: int) -> dict[str, Any]:
        req_note_ver = request.json.get("version", None)
        req_note_body = request.json.get("body", None)
        if req_note_ver is None or req_note_body is None:
            response.status = 400  # Bad request
            return {"reason": "Invalid request."}

        system_attrs = storage.get_study_system_attrs(study_id)
        if not note.version_is_incremented(system_attrs, None, req_note_ver):
            response.status = 409  # Conflict
            return {
                "reason": "The text you are editing has changed. "
                "Please copy your edits and refresh the page.",
                "note": note.get_note_from_system_attrs(system_attrs, None),
            }

        note.save_note_with_version(storage, study_id, None, req_note_ver, req_note_body)
        response.status = 204  # No content
        return {}

    @app.post("/api/studies/<study_id:int>/preference")
    @json_api_view
    def post_preference(study_id: int) -> dict[str, Any]:
        try:
            mode = request.json.get("mode", "")
            candidates = [int(d) for d in request.json.get("candidates", [])]
            clicked = int(request.json.get("clicked", -1))
        except ValueError:
            response.status = 400
            return {
                "reason": (
                    "`candidates` should be an array of integers and "
                    "`clicked` should be an integer."
                )
            }

        if clicked == -1:
            response.status = 400
            return {"reason": "`clicked` should be specified."}
        if mode != "ChooseWorst":
            response.status = 400
            return {"reason": "`mode` should be 'ChooseWorst'."}

        report_history(
            study_id,
            storage,
            NewHistory(
                mode=mode,
                candidates=candidates,
                clicked=clicked,
            ),
        )

        response.status = 204
        return {}

    @app.put("/api/studies/<study_id:int>/preference_feedback_component")
    @json_api_view
    def put_preference_feedback_component(study_id: int) -> dict[str, Any]:
        try:
            component_type = request.json.get("output_type", "")
            artifact_key = request.json.get("artifact_key", None)
        except ValueError:
            response.status = 400
            return {"reason": "invalid request."}
        if component_type not in ["note", "artifact"]:
            response.status = 400
            return {"reason": "component_type must be either 'note' or 'artifact'."}

        _register_preference_feedback_component(
            study_id=study_id,
            storage=storage,
            component_type=component_type,
            artifact_key=artifact_key,
        )
        response.status = 204
        return {}

    @app.delete("/api/studies/<study_id:int>/preference/<history_id>")
    @json_api_view
    def remove_preference(study_id: int, history_id: str) -> dict[str, Any]:
        try:
            remove_history(study_id, storage, history_id)
        except PreferenceHistoryNotFound:
            response.status = 404
            return {"reason": f"history_id={history_id} is not found"}

        response.status = 204
        return {}

    @app.post("/api/studies/<study_id:int>/preference/<history_id>")
    @json_api_view
    def restore_preference(study_id: int, history_id: str) -> dict[str, Any]:
        try:
            restore_history(study_id, storage, history_id)
        except PreferenceHistoryNotFound:
            response.status = 404
            return {"reason": f"history_id={history_id} is not found"}

        response.status = 204
        return {}

    @app.post("/api/trials/<trial_id:int>/tell")
    @json_api_view
    def tell_trial(trial_id: int) -> dict[str, Any]:
        if "state" not in request.json:
            response.status = 400  # Bad request
            return {"reason": "state must be specified."}

        try:
            state = TrialState[request.json["state"].upper()]
        except Exception:  # To catch KeyError and Exception by non str case.
            response.status = 400  # Bad request
            return {"reason": "state must be either 'Complete' or 'Fail'."}

        if state not in [TrialState.COMPLETE, TrialState.FAIL]:
            response.status = 400  # Bad request
            return {"reason": "state must be either 'Complete' or 'Fail'."}

        values = None
        if state == TrialState.COMPLETE:
            vs = request.json.get("values")
            if vs is None:
                response.status = 400  # Bad request
                return {"reason": "values attribute is required when state is 'Complete'."}
            try:
                values = [float(v) for v in vs]
            except (ValueError, TypeError):
                response.status = 400  # Bad request
                return {"reason": "values attribute must be an array of numbers"}

        storage.set_trial_state_values(trial_id, state, values)

        response.status = 204
        return {}

    @app.post("/api/trials/<trial_id:int>/user-attrs")
    @json_api_view
    def save_trial_user_attrs(trial_id: int) -> dict[str, Any]:
        user_attrs = request.json.get("user_attrs", {})
        if not user_attrs:
            response.status = 400  # Bad request
            return {"reason": "user_attrs must be specified."}

        for key, val in user_attrs.items():
            storage.set_trial_user_attr(trial_id, key, val)

        response.status = 204
        return {}

    @app.post("/api/studies/<study_id:int>/<trial_id:int>/skip")
    @json_api_view
    def skip_trial(study_id: int, trial_id: int) -> dict[str, Any]:
        try:
            system_attrs = storage.get_study_system_attrs(study_id)
        except KeyError:
            response.status = 404  # Not found
            return {"reason": f"study_id={study_id} is not found"}
        is_preferential = system_attrs.get(_SYSTEM_ATTR_PREFERENTIAL_STUDY, False)
        if not is_preferential:
            response.status = 400  # Bad request
            return {"reason": "The study is not preferential."}

        report_skip(study_id, trial_id, storage)
        response.status = 204  # No content
        return {}

    @app.put("/api/studies/<study_id:int>/<trial_id:int>/note")
    @json_api_view
    def save_trial_note(study_id: int, trial_id: int) -> dict[str, Any]:
        req_note_ver = request.json.get("version", None)
        req_note_body = request.json.get("body", None)
        if req_note_ver is None or req_note_body is None:
            response.status = 400  # Bad request
            return {"reason": "Invalid request."}

        # Store note content in study system attrs since it's always updatable.
        system_attrs = storage.get_study_system_attrs(study_id=study_id)
        if not note.version_is_incremented(system_attrs, trial_id, req_note_ver):
            response.status = 409  # Conflict
            return {
                "reason": "The text you are editing has changed. "
                "Please copy your edits and refresh the page.",
                "note": note.get_note_from_system_attrs(system_attrs, trial_id),
            }

        note.save_note_with_version(storage, study_id, trial_id, req_note_ver, req_note_body)
        response.status = 204  # No content
        return {}

    @app.get("/csv/<study_id:int>")
    def download_csv(study_id: int) -> BottleViewReturn:
        trial_ids_str = request.query.get("trial_ids", "")
        trial_ids: Optional[list[int]] = None
        if trial_ids_str:
            try:
                trial_ids = [int(tid.strip()) for tid in trial_ids_str.split(",")]
            except ValueError:
                response.status = 400  # Bad Request
                return {"reason": "Invalid trial_ids format. Expected comma-separated integers"}

        # Create a CSV file
        try:
            study_name = storage.get_study_name_from_id(study_id)
            study = optuna.load_study(storage=storage, study_name=study_name)
        except KeyError:
            response.status = 404  # Not found
            return {"reason": f"study_id={study_id} is not found"}

        if trial_ids is not None:
            trials = [t for t in study.trials if t.number in trial_ids]
            if not trials:
                response.status = 404
                return {"reason": "all specified trial_ids is not found"}
        else:
            trials = study.trials

        param_names = sorted(set(chain.from_iterable([t.params.keys() for t in trials])))
        user_attr_names = sorted(set(chain.from_iterable([t.user_attrs.keys() for t in trials])))
        param_names_header = [f"Param {x}" for x in param_names]
        user_attr_names_header = [f"UserAttribute {x}" for x in user_attr_names]
        n_objs = len(study.directions)
        if hasattr(study, "metric_names") and study.metric_names is not None:
            value_header = study.metric_names
        else:  # optuna < v3.4.0
            value_header = ["Value"] if n_objs == 1 else [f"Objective {x}" for x in range(n_objs)]
        column_names = (
            ["Number", "State"] + value_header + param_names_header + user_attr_names_header
        )

        buf = io.StringIO("")
        writer = csv.writer(buf)
        writer.writerow(column_names)
        for frozen_trial in trials:
            row = [frozen_trial.number, frozen_trial.state.name]
            row.extend(frozen_trial.values if frozen_trial.values is not None else [None] * n_objs)
            row.extend([frozen_trial.params.get(name, None) for name in param_names])
            row.extend([frozen_trial.user_attrs.get(name, None) for name in user_attr_names])
            writer.writerow(row)

        # Set response headers
        output_name = "-".join(re.sub(r'[\\/:*?"<>|]+', "", study_name).split(" "))
        response.headers["Content-Type"] = "text/csv; chatset=cp932"
        response.headers["Content-Disposition"] = f"attachment; filename={output_name}.csv"

        # Response body
        buf.seek(0)
        return buf.read()

    @app.get("/favicon.ico")
    def favicon() -> BottleViewReturn:
        use_gzip = "gzip" in request.headers["Accept-Encoding"]
        filename = "favicon.ico.gz" if use_gzip else "favicon.ico"
        return static_file(filename, IMG_DIR)

    @app.get("/static/<filename:path>")
    def send_static(filename: str) -> BottleViewReturn:
        mimetype: str | Literal[True] = True
        headers: dict[str, str] | None = None
        if not debug and "gzip" in request.headers["Accept-Encoding"]:
            gz_filename = filename.strip("/\\") + ".gz"
            if cached_path_exists(os.path.join(STATIC_DIR, gz_filename)):
                filename = gz_filename
                headers = {"Content-Encoding": "gzip"}

            mimetype_, _ = mimetypes.guess_type(filename)
            if mimetype_ is not None:
                mimetype = mimetype_
        return static_file(filename, root=STATIC_DIR, mimetype=mimetype, headers=headers)

    register_rdb_migration_route(app, storage)
    register_artifact_route(app, storage, artifact_store)
    return app


def run_server(
    storage: Union[str, BaseStorage],
    host: str = "localhost",
    port: int = 8080,
    artifact_store: Optional[ArtifactStore | ArtifactBackend] = None,
    *,
    artifact_backend: Optional[ArtifactBackend] = None,
) -> None:
    """Start running optuna-dashboard and blocks until the server terminates.

    This function uses wsgiref module which is not intended for the production
    use. If you want to run optuna-dashboard more secure and/or more fast,
    please use WSGI server like Gunicorn or uWSGI via :func:`wsgi` function.
    """
    # TODO(c-bata): Remove artifact_backend keyword argument in the future release.
    store: ArtifactStore | None = None
    if artifact_store is not None:
        store = to_artifact_store(artifact_store)
    elif artifact_backend is not None:
        warnings.warn(
            "The `artifact_backend` argument is deprecated. Please use `artifact_store` instead.",
            DeprecationWarning,
        )
        store = to_artifact_store(artifact_backend)

    app = create_app(get_storage(storage), artifact_store=store)
    run(app, host=host, port=port)


def wsgi(
    storage: Union[str, BaseStorage],
    artifact_store: Optional[ArtifactBackend | ArtifactStore] = None,
    *,
    artifact_backend: Optional[ArtifactBackend] = None,
    jupyterlab_extension_context: JupyterLabExtensionContext | None = None,
) -> WSGIApplication:
    """This function exposes WSGI interface for people who want to run on the
    production-class WSGI servers like Gunicorn or uWSGI.
    """
    # TODO(c-bata): Remove artifact_backend keyword argument in the future release.
    store: ArtifactStore | None = None
    if artifact_store is not None:
        store = to_artifact_store(artifact_store)
    elif artifact_backend is not None:
        warnings.warn(
            "The `artifact_backend` argument is deprecated. Please use `artifact_store` instead.",
            DeprecationWarning,
        )
        store = to_artifact_store(artifact_backend)

    return create_app(
        get_storage(storage),
        artifact_store=store,
        jupyterlab_extension_context=jupyterlab_extension_context,
    )



from __future__ import annotations

import logging
import threading
from typing import TYPE_CHECKING

import optuna
from optuna.importance import FanovaImportanceEvaluator
from optuna.importance import get_param_importances
from optuna.importance import MeanDecreaseImpurityImportanceEvaluator
from optuna.storages import BaseStorage
from optuna.study import Study
from optuna.trial import FrozenTrial
from optuna.trial import TrialState
from optuna_dashboard._inmemory_cache import get_cached_extra_study_property
from optuna_dashboard._inmemory_cache import InMemoryCache


_logger = logging.getLogger(__name__)


try:
    from optuna.importance import PedAnovaImportanceEvaluator  # type: ignore[attr-defined]
except ImportError:
    PedAnovaImportanceEvaluator = None  # type: ignore


if TYPE_CHECKING:
    from typing import Callable
    from typing import Optional
    from typing import TypedDict

    ImportanceType = TypedDict(
        "ImportanceType",
        {
            "name": str,
            "importance": float,
            "distribution": str,
        },
    )

param_importance_cache_lock = threading.Lock()
# { "{study_id}:{objective_id}" : (n_completed_trials, importance) }
param_importance_cache: dict[str, tuple[int, list[ImportanceType]]] = {}


class StudyWrapper(Study):
    def __init__(
        self, storage: BaseStorage, study_id: int, cached_trials: list[FrozenTrial]
    ) -> None:
        study_name = storage.get_study_name_from_id(study_id)
        super().__init__(study_name=study_name, storage=storage)
        self._cached_trials = cached_trials

    @property
    def trials(self) -> list[FrozenTrial]:
        return self._cached_trials


def _get_param_importances(
    study: optuna.Study,
    *,
    target: Optional[Callable[[FrozenTrial], float]] = None,
    evaluator: str,
) -> dict[str, float]:
    if evaluator == "ped_anova":
        if PedAnovaImportanceEvaluator is not None:
            # TODO(nabenabe0928): We might want to pass baseline_quantile
            #                     as an argument in the future.
            return get_param_importances(
                study, target=target, evaluator=PedAnovaImportanceEvaluator()
            )
    elif evaluator == "mean_decrease_impurity":
        return get_param_importances(
            study, target=target, evaluator=MeanDecreaseImpurityImportanceEvaluator()
        )

    return get_param_importances(
        study,
        target=target,
        evaluator=FanovaImportanceEvaluator(),
    )


def get_param_importance_from_trials_cache(
    inmemory_cache: InMemoryCache,
    storage: BaseStorage,
    study_id: int,
    objective_id: int,
    evaluator: str,
    trials: list[FrozenTrial],
) -> list[ImportanceType]:
    n_completed_trials = len([t for t in trials if t.state == TrialState.COMPLETE])
    if n_completed_trials <= 1:
        return []

    cache_key = f"{study_id}:{objective_id}:{evaluator}"
    with param_importance_cache_lock:
        cache_n_trial, cache_importance = param_importance_cache.get(cache_key, (0, []))
        if n_completed_trials == cache_n_trial:
            return cache_importance

        study = StudyWrapper(storage, study_id, trials)
        try:
            importance = _get_param_importances(
                study,
                target=lambda t: t.values[objective_id],
                evaluator=evaluator,
            )
        except RuntimeError:
            # RuntimeError("Encountered zero total variance in all trees.") may be raised
            # when all objective values are same.
            _, union_search_space, _, _ = get_cached_extra_study_property(
                inmemory_cache, study_id, trials
            )
            importance_value = 1 / len(union_search_space)
            importance = {
                param_name: importance_value for param_name, distribution in union_search_space
            }
        converted = convert_to_importance_type(importance, trials)
        param_importance_cache[cache_key] = (n_completed_trials, converted)
    return converted


def convert_to_importance_type(
    importance: dict[str, float], trials: list[FrozenTrial]
) -> list[ImportanceType]:
    return [
        {
            "name": name,
            "importance": importance,
            "distribution": get_distribution_name(name, trials),
        }
        for name, importance in importance.items()
    ]


def get_distribution_name(param_name: str, trials: list[FrozenTrial]) -> str:
    for trial in trials:
        if param_name in trial.distributions:
            return trial.distributions[param_name].__class__.__name__
    assert False, "Must not reach here."
